import logging

import core.input
import core.decorators

import util.store
import util.format

log = logging.getLogger(__name__)


class Widget(util.store.Store, core.input.Object):
    def __init__(self, full_text="", name=None, widget_id=None):
        super(Widget, self).__init__()
        self.__full_text = full_text
        self.module = None
        self.name = name
        self.id = widget_id or self.id

    @property
    def module(self):
        return self.__module

    @module.setter
    def module(self, module):
        self.__module = module

        if self.index() < 0:
            return

        if module:
            custom_ids = util.format.aslist(module.parameter("id"))
            if len(custom_ids) > self.index():
                self.id = custom_ids[self.index()]
            if util.format.asbool(module.parameter("scrolling", False)) == True:
                if callable(self.__full_text):
                    self.__full_text = core.decorators.scrollable(
                        self.__full_text.__func__
                    ).__get__(module)
                else:
                    log.warning("unable to make scrollable: {}".format(module.name))

    def index(self):
        if not self.module:
            return 0

        idx = 0
        for w in self.module.widgets():
            if w.id == self.id:
                return idx
            idx = idx + 1
        return -1  # not found

    def theme(self, attribute):
        attr = "theme.{}".format(attribute)
        if self.module:
            param = util.format.aslist(self.module.parameter(attr))
            if param and len(param) > self.index():
                return param[self.index()]
        return self.get(attr)

    def full_text(self, value=None):
        if value:
            self.__full_text = value
        else:
            if callable(self.__full_text):
                return self.__full_text(self)
        return self.__full_text

    def state(self):
        rv = []
        if self.get("state", None):
            tmp = self.get("state")
            rv = tmp[:] if isinstance(tmp, list) else [tmp]
        if self.module:
            tmp = self.module.state(self)
            rv.extend(tmp if isinstance(tmp, list) else [tmp])
        return rv


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903
# -*- coding: utf-8 -*-

"""Displays information about the current song in mpd.

Requires the following executable:
    * mpc

Parameters:
    * mpd.format: Format string for the song information.

      Supported tags (see `man mpc` for additional information)

         * {name}
         * {artist}
         * {album}
         * {albumartist}
         * {comment}
         * {composer}
         * {date}
         * {originaldate}
         * {disc}
         * {genre}
         * {performer}
         * {title}
         * {track}
         * {time}
         * {file}
         * {id}
         * {prio}
         * {mtime}
         * {mdate}

      Additional tags:

         * {position} - position of currently playing song
           not to be confused with %position% mpc tag
         * {duration} - duration of currently playing song
         * {file1} - song file name without path prefix
           if {file} = '/foo/bar.baz', then {file1} = 'bar.baz'
         * {file2} - song file name without path prefix and extension suffix
           if {file} = '/foo/bar.baz', then {file2} = 'bar'

    * mpd.host: MPD host to connect to. (mpc behaviour by default)
    * mpd.layout: Space-separated list of widgets to add. Possible widgets are the buttons/toggles mpd.prev, mpd.next, mpd.shuffle and mpd.repeat, and the main display with play/pause function mpd.main.

contributed by `alrayyes <https://github.com/alrayyes>`_ - many thanks!
"""

from collections import defaultdict

import string
import os

import core.module
import core.input
import core.decorators

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self._layout = self.parameter(
            "layout", "mpd.prev mpd.main mpd.next mpd.shuffle mpd.repeat"
        )

        self._fmt = self.parameter("format", "{artist} - {title} {position}/{duration}")
        self._status = None
        self._shuffle = False
        self._repeat = False
        self._tags = defaultdict(lambda: "")

        if not self.parameter("host"):
            self._hostcmd = ""
        else:
            self._hostcmd = " -h " + self.parameter("host")

        # Create widgets
        widget_map = {}
        for widget_name in self._layout.split():
            widget = self.add_widget(name=widget_name)

            if widget_name == "mpd.prev":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "mpc prev" + self._hostcmd,
                }
            elif widget_name == "mpd.main":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "mpc toggle" + self._hostcmd,
                }
                widget.full_text(self.description)
            elif widget_name == "mpd.next":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "mpc next" + self._hostcmd,
                }
            elif widget_name == "mpd.shuffle":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "mpc random" + self._hostcmd,
                }
            elif widget_name == "mpd.repeat":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "mpc repeat" + self._hostcmd,
                }
            else:
                raise KeyError(
                    "The mpd module does not support a {widget_name!r} widget".format(
                        widget_name=widget_name
                    )
                )

        # Register input callbacks
        for widget, callback_options in widget_map.items():
            core.input.register(widget, **callback_options)

    def hidden(self):
        return self._status is None

    @core.decorators.scrollable
    def description(self, widget):
        return string.Formatter().vformat(self._fmt, (), self._tags)

    def update(self):
        self._load_song()

    def state(self, widget):
        if widget.name == "mpd.shuffle":
            return "shuffle-on" if self._shuffle else "shuffle-off"
        if widget.name == "mpd.repeat":
            return "repeat-on" if self._repeat else "repeat-off"
        if widget.name == "mpd.prev":
            return "prev"
        if widget.name == "mpd.next":
            return "next"
        return self._status

    def _load_song(self):
        info = ""
        tags = [
            "name",
            "artist",
            "album",
            "albumartist",
            "comment",
            "composer",
            "date",
            "originaldate",
            "disc",
            "genre",
            "performer",
            "title",
            "track",
            "time",
            "file",
            "id",
            "prio",
            "mtime",
            "mdate",
        ]
        joinedtags = "\n".join(["tag {0} %{0}%".format(tag) for tag in tags])
        info = util.cli.execute(
            'mpc -f "{}"{}'.format(joinedtags, self._hostcmd), ignore_errors=True
        )

        self._tags = defaultdict(lambda: "")
        self._status = None
        for line in info.split("\n"):
            if line.startswith("[playing]"):
                self._status = "playing"
            elif line.startswith("[paused]"):
                self._status = "paused"

            if line.startswith("["):
                timer = line.split()[2]
                position = timer.split("/")[0]
                dur = timer.split("/")[1]
                duration = dur.split(" ")[0]
                self._tags.update({"position": position})
                self._tags.update({"duration": duration})

            if line.startswith("volume"):
                value = line.split("   ", 2)[1:]
                for option in value:
                    if option.startswith("repeat: on"):
                        self._repeat = True
                    elif option.startswith("repeat: off"):
                        self._repeat = False
                    elif option.startswith("random: on"):
                        self._shuffle = True
                    elif option.startswith("random: off"):
                        self._shuffle = False
            if line.startswith("tag"):
                key, value = line.split(" ", 2)[1:]
                self._tags.update({key: value})
                if key == "file":
                    self._tags.update({"file1": os.path.basename(value)})
                    self._tags.update(
                        {"file2": os.path.splitext(os.path.basename(value))[0]}
                    )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: UTF-8 -*-
# pylint: disable=C0111,R0903

"""Displays sensor temperature

Parameters:
    * sensors.path: path to temperature file (default /sys/class/thermal/thermal_zone0/temp).
    * sensors.json: if set to 'true', interpret sensors.path as JSON 'path' in the output
      of 'sensors -j' (i.e. <key1>/<key2>/.../<value>), for example, path could
      be: 'coretemp-isa-00000/Core 0/temp1_input' (defaults to 'false')
    * sensors.match: (fallback) Line to match against output of 'sensors -u' (default: temp1_input)
    * sensors.match_pattern: (fallback) Line to match against before temperature is read (no default)
    * sensors.match_number: (fallback) which of the matches you want (default -1: last match).
    * sensors.show_freq: whether to show CPU frequency. (default: true)


contributed by `mijoharas <https://github.com/mijoharas>`_ - many thanks!
"""

import re
import json
import logging

log = logging.getLogger(__name__)

import core.module
import core.widget
import core.input

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.temperature))

        self._temperature = "unknown"
        self._mhz = "n/a"
        self._match_number = int(self.parameter("match_number", "-1"))
        self._match_pattern = self.parameter("match_pattern", None)
        self._pattern = re.compile(
            r"^\s*{}:\s*([\d.]+)$".format(self.parameter("match", "temp1_input")),
            re.MULTILINE,
        )
        self._json = util.format.asbool(self.parameter("json", False))
        self._freq = util.format.asbool(self.parameter("show_freq", True))
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd="xsensors")
        self.determine_method()

    def determine_method(self):
        if self.parameter("path") != None and self._json == False:
            self.use_sensors = False  # use thermal zone
        else:
            # try to use output of sensors -u
            try:
                output = util.cli.execute("sensors -u")
                self.use_sensors = True
                log.debug("Sensors command available")
            except FileNotFoundError as e:
                log.info(
                    "Sensors command not available, using /sys/class/thermal/thermal_zone*/"
                )
                self.use_sensors = False

    def _get_temp_from_sensors(self):
        if self._json == True:
            try:
                output = json.loads(util.cli.execute("sensors -j"))
                for key in self.parameter("path").split("/"):
                    output = output[key]
                return int(float(output))
            except Exception as e:
                logging.error("unable to read sensors: {}".format(str(e)))
                return "unknown"
        else:
            output = util.cli.execute("sensors -u")
            if self._match_pattern:
                temp_pattern = self.parameter("match", "temp1_input")
                match = re.search(
                    r"{}.+{}:\s*([\d.]+)$".format(self._match_pattern, temp_pattern),
                    output.replace("\n", ""),
                )
                if match:
                    return int(float(match.group(1)))
                else:
                    return "unknown"
            match = self._pattern.findall(output)
            if match:
                return int(float(match[self._match_number]))
        return "unknown"

    def get_temp(self):
        if self.use_sensors:
            temperature = self._get_temp_from_sensors()
            log.debug("Retrieve temperature from sensors -u")
        else:
            try:
                temperature = open(
                    self.parameter("path", "/sys/class/thermal/thermal_zone0/temp")
                ).read()[:2]
                log.debug("retrieved temperature from /sys/class/")
                # TODO: Iterate through all thermal zones to determine the correct one and use its value
                # https://unix.stackexchange.com/questions/304845/discrepancy-between-number-of-cores-and-thermal-zones-in-sys-class-thermal

            except IOError:
                temperature = "unknown"
                log.info("Can not determine temperature, please install lm-sensors")

        return temperature

    def get_mhz(self):
        mhz = None
        try:
            output = open(
                "/sys/devices/system/cpu/cpufreq/policy0/scaling_cur_freq"
            ).read()
            mhz = int(float(output) / 1000.0)
        except:
            output = open("/proc/cpuinfo").read()
            m = re.search(r"cpu MHz\s+:\s+(\d+)", output)
            if m:
                mhz = int(m.group(1))
            else:
                m = re.search(r"BogoMIPS\s+:\s+(\d+)", output)
                if m:
                    return "{} BogoMIPS".format(int(m.group(1)))
        if not mhz:
            return "n/a"

        if mhz < 1000:
            return "{} MHz".format(mhz)
        else:
            return "{:0.01f} GHz".format(float(mhz) / 1000.0)

    def temperature(self, _):
        if self._freq:
            return "{}°c @ {}".format(self._temperature, self._mhz)
        else:
            return "{}°c".format(self._temperature)

    def update(self):
        self._temperature = self.get_temp()
        if self._freq:
            self._mhz = self.get_mhz()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import os
import io
import json
import logging
import copy
import glob

import core.event
import util.algorithm

log = logging.getLogger(__name__)

THEME_BASE_DIR = os.path.dirname(os.path.realpath(__file__))
PATHS = [
    ".",
    os.path.join(THEME_BASE_DIR, "../../themes"),
    os.path.expanduser("~/.config/bumblebee-status/themes"),
    os.path.expanduser("~/.local/share/bumblebee-status/themes"),  # PIP
]


def themes():
    themes_dict = {}

    for path in PATHS:
        for filename in glob.iglob("{}/*.json".format(path)):
            if "test" not in filename:
                themes_dict[os.path.basename(filename).replace(".json", "")] = 1
    result = list(themes_dict.keys())
    result.sort()
    return result


def merge_replace(value, new_value, key):
    if not isinstance(value, dict):
        return new_value
    if isinstance(new_value, dict):
        return util.algorithm.merge(new_value, value)
    # right now, merging needs explicit pango support :(
    if "pango" in value:
        value["pango"]["full_text"] = new_value
    return value


class Theme(object):
    def __init__(self, name="default", iconset="auto", raw_data=None):
        self.name = name
        self.__widget_count = 0
        self.__previous = {}
        self.__current = {}
        self.__keywords = {}
        self.__value_idx = {}
        self.__data = raw_data if raw_data else self.load(name)

        for icons in self.__data.get("icons", []):
            self.__data = util.algorithm.merge(self.load(icons, "icons"), self.__data)
        if iconset != "auto":
            self.__data = util.algorithm.merge(self.load(iconset, "icons"), self.__data)
        for colors in self.__data.get("colors", []):
            util.algorithm.merge(self.__keywords, self.load_keywords(colors))

        core.event.register("draw", self.__start)
        core.event.register("next-widget", self.__next_widget)

    def keywords(self):
        return self.__keywords

    def color(self, name, default=None):
        return self.keywords().get(name, default)

    def load(self, name, subdir=""):
        if isinstance(name, dict):
            return name  # support plain data
        for path in PATHS:
            theme_file = os.path.join(path, subdir, "{}.json".format(name))
            result = self.__load_json(theme_file)
            if result != {}:
                return result
        raise RuntimeError("unable to find theme {}".format(name))

    def __load_json(self, filename):
        filename = os.path.expanduser(filename)
        if not os.path.isfile(filename):
            return {}
        with io.open(filename) as data:
            return json.load(data)

    def load_keywords(self, name):
        try:
            if isinstance(name, dict):
                return name
            if name.lower() == "wal":
                wal = self.__load_json("~/.cache/wal/colors.json")
                result = {}
                for field in ["special", "colors"]:
                    for key in wal.get(field, {}):
                        result[key] = wal[field][key]
                return result
        except Exception as e:
            log.error("failed to load colors: {}", e)

    def __start(self):
        self.__widget_count = 0
        self.__current.clear()
        self.__previous.clear()

        for key, value in self.__value_idx.items():
            self.__value_idx[key] = value + 1

    def __next_widget(self):
        self.__widget_count = self.__widget_count + 1
        self.__previous = dict(self.__current)
        self.__current.clear()

    def get(self, key, widget=None, default=None):
        if not widget:
            widget = core.widget.Widget("")
        # special handling
        if widget == "previous":
            return self.__previous.get(key, None)

        value = default

        for option in ["defaults", "cycle"]:
            if option in self.__data:
                tmp = self.__data[option]
                if isinstance(tmp, list):
                    tmp = tmp[self.__widget_count % len(tmp)]
                value = merge_replace(value, tmp.get(key, value), key)

        if isinstance(value, dict):
            value = copy.deepcopy(value)

        value = merge_replace(value, self.__data.get(key, value), key)

        if widget.module:
            value = merge_replace(
                value, self.get(widget.module.name, None, {}).get(key, value), key
            )
            value = merge_replace(
                value, self.get(widget.module.id, None, {}).get(key, value), key
            )

        if not key in widget.state():
            for state in widget.state():
                theme = self.get(state, widget, {})
                value = merge_replace(value, theme.get(key, value), key)

        if not type(value) in (list, dict):
            value = self.__keywords.get(value, value)

        if isinstance(value, list):
            idx = self.__value_idx.get("{}::{}".format(widget.id, key), 0) % len(value)
            self.__value_idx["{}::{}".format(widget.id, key)] = idx
            widget.set(key, idx)
            value = value[idx]
        self.__current[key] = value
        return value


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# UPowerManger Class Copyright (C) 2017 Oscar Svensson (wogscpar) under MIT licence from upower-python

"""Displays battery status, remaining percentage and charging information.

Parameters:
    * battery-upower.warning      : Warning threshold in % of remaining charge (defaults to 20)
    * battery-upower.critical     : Critical threshold in % of remaining charge (defaults to 10)
    * battery-upower.showremaining : If set to true (default) shows the remaining time until the batteries are completely discharged

contributed by `martindoublem <https://github.com/martindoublem>`_ - many thanks!
"""

import dbus
import logging

import core.module
import core.widget
import core.input

import util.format


class UPowerManager:
    def __init__(self):
        self.UPOWER_NAME = "org.freedesktop.UPower"
        self.UPOWER_PATH = "/org/freedesktop/UPower"

        self.DBUS_PROPERTIES = "org.freedesktop.DBus.Properties"
        self.bus = dbus.SystemBus()

    def detect_devices(self):
        upower_proxy = self.bus.get_object(self.UPOWER_NAME, self.UPOWER_PATH)
        upower_interface = dbus.Interface(upower_proxy, self.UPOWER_NAME)

        devices = upower_interface.EnumerateDevices()
        return devices

    def get_display_device(self):
        upower_proxy = self.bus.get_object(self.UPOWER_NAME, self.UPOWER_PATH)
        upower_interface = dbus.Interface(upower_proxy, self.UPOWER_NAME)

        dispdev = upower_interface.GetDisplayDevice()
        return dispdev

    def get_critical_action(self):
        upower_proxy = self.bus.get_object(self.UPOWER_NAME, self.UPOWER_PATH)
        upower_interface = dbus.Interface(upower_proxy, self.UPOWER_NAME)

        critical_action = upower_interface.GetCriticalAction()
        return critical_action

    def get_device_percentage(self, battery):
        battery_proxy = self.bus.get_object(self.UPOWER_NAME, battery)
        battery_proxy_interface = dbus.Interface(battery_proxy, self.DBUS_PROPERTIES)

        return battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Percentage")

    def get_full_device_information(self, battery):
        battery_proxy = self.bus.get_object(self.UPOWER_NAME, battery)
        battery_proxy_interface = dbus.Interface(battery_proxy, self.DBUS_PROPERTIES)

        hasHistory = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "HasHistory"
        )
        hasStatistics = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "HasStatistics"
        )
        isPresent = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "IsPresent"
        )
        isRechargable = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "IsRechargeable"
        )
        online = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Online")
        powersupply = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "PowerSupply"
        )
        capacity = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Capacity")
        energy = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Energy")
        energyempty = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "EnergyEmpty"
        )
        energyfull = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "EnergyFull"
        )
        energyfulldesign = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "EnergyFullDesign"
        )
        energyrate = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "EnergyRate"
        )
        luminosity = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "Luminosity"
        )
        percentage = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "Percentage"
        )
        temperature = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "Temperature"
        )
        voltage = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Voltage")
        timetoempty = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "TimeToEmpty"
        )
        timetofull = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "TimeToFull"
        )
        iconname = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "IconName")
        model = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Model")
        nativepath = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "NativePath"
        )
        serial = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Serial")
        vendor = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Vendor")
        state = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "State")
        technology = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "Technology"
        )
        battype = battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "Type")
        warninglevel = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "WarningLevel"
        )
        updatetime = battery_proxy_interface.Get(
            self.UPOWER_NAME + ".Device", "UpdateTime"
        )

        information_table = {
            "HasHistory": hasHistory,
            "HasStatistics": hasStatistics,
            "IsPresent": isPresent,
            "IsRechargeable": isRechargable,
            "Online": online,
            "PowerSupply": powersupply,
            "Capacity": capacity,
            "Energy": energy,
            "EnergyEmpty": energyempty,
            "EnergyFull": energyfull,
            "EnergyFullDesign": energyfulldesign,
            "EnergyRate": energyrate,
            "Luminosity": luminosity,
            "Percentage": percentage,
            "Temperature": temperature,
            "Voltage": voltage,
            "TimeToEmpty": timetoempty,
            "TimeToFull": timetofull,
            "IconName": iconname,
            "Model": model,
            "NativePath": nativepath,
            "Serial": serial,
            "Vendor": vendor,
            "State": state,
            "Technology": technology,
            "Type": battype,
            "WarningLevel": warninglevel,
            "UpdateTime": updatetime,
        }

        return information_table

    def is_lid_present(self):
        upower_proxy = self.bus.get_object(self.UPOWER_NAME, self.UPOWER_PATH)
        upower_interface = dbus.Interface(upower_proxy, self.DBUS_PROPERTIES)

        is_lid_present = bool(upower_interface.Get(self.UPOWER_NAME, "LidIsPresent"))
        return is_lid_present

    def is_lid_closed(self):
        upower_proxy = self.bus.get_object(self.UPOWER_NAME, self.UPOWER_PATH)
        upower_interface = dbus.Interface(upower_proxy, self.DBUS_PROPERTIES)

        is_lid_closed = bool(upower_interface.Get(self.UPOWER_NAME, "LidIsClosed"))
        return is_lid_closed

    def on_battery(self):
        upower_proxy = self.bus.get_object(self.UPOWER_NAME, self.UPOWER_PATH)
        upower_interface = dbus.Interface(upower_proxy, self.DBUS_PROPERTIES)

        on_battery = bool(upower_interface.Get(self.UPOWER_NAME, "OnBattery"))
        return on_battery

    def has_wakeup_capabilities(self):
        upower_proxy = self.bus.get_object(
            self.UPOWER_NAME, self.UPOWER_PATH + "/Wakeups"
        )
        upower_interface = dbus.Interface(upower_proxy, self.DBUS_PROPERTIES)

        has_wakeup_capabilities = bool(
            upower_interface.Get(self.UPOWER_NAME + ".Wakeups", "HasCapability")
        )
        return has_wakeup_capabilities

    def get_wakeups_data(self):
        upower_proxy = self.bus.get_object(
            self.UPOWER_NAME, self.UPOWER_PATH + "/Wakeups"
        )
        upower_interface = dbus.Interface(upower_proxy, self.UPOWER_NAME + ".Wakeups")

        data = upower_interface.GetData()
        return data

    def get_wakeups_total(self):
        upower_proxy = self.bus.get_object(
            self.UPOWER_NAME, self.UPOWER_PATH + "/Wakeups"
        )
        upower_interface = dbus.Interface(upower_proxy, self.UPOWER_NAME + ".Wakeups")

        data = upower_interface.GetTotal()
        return data

    def is_loading(self, battery):
        battery_proxy = self.bus.get_object(self.UPOWER_NAME, battery)
        battery_proxy_interface = dbus.Interface(battery_proxy, self.DBUS_PROPERTIES)

        state = int(battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "State"))

        if state == 1:
            return True
        else:
            return False

    def get_state(self, battery):
        battery_proxy = self.bus.get_object(self.UPOWER_NAME, battery)
        battery_proxy_interface = dbus.Interface(battery_proxy, self.DBUS_PROPERTIES)

        state = int(battery_proxy_interface.Get(self.UPOWER_NAME + ".Device", "State"))

        if state == 0:
            return "Unknown"
        elif state == 1:
            return "Loading"
        elif state == 2:
            return "Discharging"
        elif state == 3:
            return "Empty"
        elif state == 4:
            return "Fully charged"
        elif state == 5:
            return "Pending charge"
        elif state == 6:
            return "Pending discharge"


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.capacity))

        try:
            self.power = UPowerManager()
            self.device = self.power.get_display_device()
        except Exception as e:
            logging.exception("unable to get battery display device: {}".format(str(e)))
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="gnome-power-statistics"
        )

        self._showremaining = util.format.asbool(self.parameter("showremaining", True))

    def capacity(self, widget):
        widget.set("capacity", -1)
        widget.set("ac", False)
        output = "n/a"
        try:
            capacity = int(self.power.get_device_percentage(self.device))
            capacity = capacity if capacity < 100 else 100
            widget.set("capacity", capacity)
            output = "{}%".format(capacity)
            widget.set("theme.minwidth", "100%")
        except Exception as e:
            logging.exception("unable to get battery capacity: {}".format(str(e)))

        if self._showremaining:
            try:
                p = self.power  # an alias to make each line of code shorter
                proxy = p.bus.get_object(p.UPOWER_NAME, self.device)
                interface = dbus.Interface(proxy, p.DBUS_PROPERTIES)
                state = int(interface.Get(p.UPOWER_NAME + ".Device", "State"))
                # state: 1 => charging, 2 => discharging, other => don't care
                remain = int(
                    interface.Get(
                        p.UPOWER_NAME + ".Device",
                        ["TimeToFull", "TimeToEmpty"][state - 1],
                    )
                )
                remain = util.format.duration(remain, compact=True, unit=True)
                output = "{} {}".format(output, remain)
            except IndexError:
                pass
            except Exception as e:
                logging.exception(
                    "unable to get battery remaining time: {}".format(str(e))
                )

        return output

    def state(self, widget):
        state = []
        capacity = widget.get("capacity", -1)
        if capacity < 0:
            return ["critical", "unknown"]

        if capacity < int(self.parameter("critical", 10)):
            state.append("critical")
        elif capacity < int(self.parameter("warning", 20)):
            state.append("warning")

        if widget.get("ac"):
            state.append("AC")
        else:
            charge = "Unknown"
            try:
                charge = self.power.get_state(self.device)
            except Exception as e:
                logging.exception("unable to get charge value: {}".format(str(e)))
            if charge == "Discharging":
                state.append(
                    "discharging-{}".format(
                        min([10, 25, 50, 80, 100], key=lambda i: abs(i - capacity))
                    )
                )
            elif charge == "Unknown":
                state.append(
                    "unknown-{}".format(
                        min([10, 25, 50, 80, 100], key=lambda i: abs(i - capacity))
                    )
                )
            else:
                if capacity > 95:
                    state.append("charged")
                else:
                    state.append("charging")
        return state


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the current song being played

Requires the following library:
    * python-dbus

Parameters:
    * deezer.format:   Format string (defaults to '{artist} - {title}')
      Available values are: {album}, {title}, {artist}, {trackNumber}, {playbackStatus}
    * deezer.previous: Change binding for previous song (default is left click)
    * deezer.next:     Change binding for next song (default is right click)
    * deezer.pause:    Change binding for toggling pause (default is middle click)

    Available options for deezer.previous, deezer.next and deezer.pause are:
        LEFT_CLICK, RIGHT_CLICK, MIDDLE_CLICK, SCROLL_UP, SCROLL_DOWN

contributed by `wwmoraes <https://github.com/wwmoraes>`_ - many thanks!
"""

import dbus

import core.module
import core.widget
import core.input


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.deezer))

        buttons = {
            "LEFT_CLICK": core.input.LEFT_MOUSE,
            "RIGHT_CLICK": core.input.RIGHT_MOUSE,
            "MIDDLE_CLICK": core.input.MIDDLE_MOUSE,
            "SCROLL_UP": core.input.WHEEL_UP,
            "SCROLL_DOWN": core.input.WHEEL_DOWN,
        }

        self._song = ""
        self._format = self.parameter("format", "{artist} - {title}")
        prev_button = self.parameter("previous", "LEFT_CLICK")
        next_button = self.parameter("next", "RIGHT_CLICK")
        pause_button = self.parameter("pause", "MIDDLE_CLICK")

        cmd = "dbus-send --session --type=method_call --dest=org.mpris.MediaPlayer2.deezer \
                /org/mpris/MediaPlayer2 org.mpris.MediaPlayer2.Player."
        core.input.register(self, button=buttons[prev_button], cmd=cmd + "Previous")
        core.input.register(self, button=buttons[next_button], cmd=cmd + "Next")
        core.input.register(self, button=buttons[pause_button], cmd=cmd + "PlayPause")

    def deezer(self, widget):
        return str(self._song)

    def hidden(self):
        return str(self._song) == ""

    def update(self):
        try:
            bus = dbus.SessionBus()
            deezer = bus.get_object(
                "org.mpris.MediaPlayer2.deezer", "/org/mpris/MediaPlayer2"
            )
            deezer_iface = dbus.Interface(deezer, "org.freedesktop.DBus.Properties")
            props = deezer_iface.Get("org.mpris.MediaPlayer2.Player", "Metadata")
            playback_status = str(
                deezer_iface.Get("org.mpris.MediaPlayer2.Player", "PlaybackStatus")
            )
            self._song = self._format.format(
                album=str(props.get("xesam:album")),
                title=str(props.get("xesam:title")),
                artist=",".join(props.get("xesam:artist")),
                trackNumber=str(props.get("xesam:trackNumber")),
                playbackStatus="\u25B6"
                if playback_status == "Playing"
                else "\u258D\u258D"
                if playback_status == "Paused"
                else "",
            )
        except Exception:
            self._song = ""


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: utf-8 -*-

"""Displays the number of docker containers running

Requires the following python packages:
    * docker

contributed by `jlopezzarza <https://github.com/jlopezzarza>`_ - many thanks!
"""

import docker

from requests.exceptions import ConnectionError

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(seconds=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.docker_info))
        self.__info = ""

    def state(self, widget):
        state = []
        if self.__info == "OK - 0":
            state.append("warning")
        elif self.__info in ["n/a", "off"]:
            state.append("critical")
        return state

    def docker_info(self, widget):
        try:
            cli = docker.DockerClient(base_url="unix://var/run/docker.sock")
            cli.ping()
            self.__info = "OK - {}".format(
                len(cli.containers.list(filters={"status": "running"}))
            )
        except ConnectionError:
            self.__info = "off"
        except Exception:
            self.__info = "n/a"
        return self.__info


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import os
import ast

from configparser import RawConfigParser

import sys
import glob
import textwrap
import argparse
import logging

import core.theme

import util.store
import util.format

import modules.core
import modules.contrib

log = logging.getLogger(__name__)

MODULE_HELP = "Specify a space-separated list of modules to load. The order of the list determines their order in the i3bar (from left to right). Use <module>:<alias> to provide an alias in case you want to load the same module multiple times, but specify different parameters."
PARAMETER_HELP = (
    "Provide configuration parameters in the form of <module>.<key>=<value>"
)
THEME_HELP = "Specify the theme to use for drawing modules"


def all_modules():
    """Returns a list of all available modules (either core or contrib)

    :return: list of modules
    :rtype: list of strings
    """
    result = {}

    for path in [modules.core.__file__, modules.contrib.__file__]:
        path = os.path.dirname(path)
        for mod in glob.iglob("{}/*.py".format(path)):
            result[os.path.basename(mod).replace(".py", "")] = 1

    res = list(result.keys())
    res.sort()
    return res


class print_usage(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        argparse.Action.__init__(self, option_strings, dest, nargs, **kwargs)
        self._indent = " " * 2

    def __call__(self, parser, namespace, value, option_string=None):
        if value == "modules":
            self._args = namespace
            self._format = "plain"
            self.print_modules()
        elif value == "modules-rst":
            self._args = namespace
            self._format = "rst"
            self.print_modules()
        elif value == "themes":
            self.print_themes()
        sys.exit(0)

    def print_themes(self):
        print(", ".join(core.theme.themes()))

    def print_modules(self):
        basepath = os.path.abspath(
            os.path.join(os.path.dirname(os.path.realpath(__file__)), "..")
        )

        rst = {}
        for m in all_modules():
            try:
                module_type = "core"
                filename = os.path.join(basepath, "modules", "core", "{}.py".format(m))
                if not os.path.exists(filename):
                    filename = os.path.join(
                        basepath, "modules", "contrib", "{}.py".format(m)
                    )
                    module_type = "contrib"
                if not os.path.exists(filename):
                    log.warning("module {} not found".format(m))
                    continue

                doc = None
                with open(filename) as f:
                    tree = ast.parse(f.read())
                    doc = ast.get_docstring(tree)

                if not doc:
                    log.warning("failed to find docstring for {}".format(m))
                    continue
                if self._format == "rst":
                    if os.path.exists(
                        os.path.join(basepath, "..", "screenshots", "{}.png".format(m))
                    ):
                        doc = "{}\n\n.. image:: ../screenshots/{}.png".format(doc, m)

                    rst[module_type] = rst.get(module_type, [])
                    rst[module_type].append({"module": m, "content": doc})
                else:
                    print(
                        textwrap.fill(
                            "{}:".format(m),
                            80,
                            initial_indent=self._indent * 2,
                            subsequent_indent=self._indent * 2,
                        )
                    )
                    for line in doc.split("\n"):
                        print(
                            textwrap.fill(
                                line,
                                80,
                                initial_indent=self._indent * 3,
                                subsequent_indent=self._indent * 6,
                            )
                        )
            except Exception as e:
                log.warning(e)

        if self._format == "rst":
            print("List of modules\n===============")
            for k in ["core", "contrib"]:
                print("\n{}\n{}\n".format(k, "-" * len(k)))
                for mod in rst[k]:
                    print("\n{}\n{}\n".format(mod["module"], "~" * len(mod["module"])))
                    print(mod["content"])


class Config(util.store.Store):
    """Represents the configuration of bumblebee-status (either via config file or via CLI)

    :param args: The arguments passed via the commandline
    """

    def __init__(self, args):
        super(Config, self).__init__()

        parser = argparse.ArgumentParser(
            description="bumblebee-status is a modular, theme-able status line generator for the i3 window manager. https://github.com/tobi-wan-kenobi/bumblebee-status/wiki"
        )
        parser.add_argument(
            "-m", "--modules", nargs="+", action="append", default=[], help=MODULE_HELP
        )
        parser.add_argument(
            "-p",
            "--parameters",
            nargs="+",
            action="append",
            default=[],
            help=PARAMETER_HELP,
        )
        parser.add_argument("-t", "--theme", default="default", help=THEME_HELP)
        parser.add_argument(
            "-i",
            "--iconset",
            default="auto",
            help="Specify the name of an iconset to use (overrides theme default)",
        )
        parser.add_argument(
            "-a",
            "--autohide",
            nargs="+",
            default=[],
            help="Specify a list of modules to hide when not in warning/error state",
        )
        parser.add_argument(
            "-d", "--debug", action="store_true", help="Add debug fields to i3 output"
        )
        parser.add_argument(
            "-f",
            "--logfile",
            help="destination for the debug log file, if -d|--debug is specified; defaults to stderr",
        )
        parser.add_argument(
            "-r",
            "--right-to-left",
            action="store_true",
            help="Draw widgets from right to left, rather than left to right (which is the default)",
        )
        parser.add_argument(
            "-l",
            "--list",
            choices=["modules", "themes", "modules-rst"],
            help="Display a list of available themes or available modules, along with their parameters",
            action=print_usage,
        )

        self.__args = parser.parse_args(args)

        for cfg in [
            "~/.bumblebee-status.conf",
            "~/.config/bumblebee-status.conf",
            "~/.config/bumblebee-status/config",
        ]:
            cfg = os.path.expanduser(cfg)
            self.load_config(cfg)

        parameters = [item for sub in self.__args.parameters for item in sub]
        for param in parameters:
            if not "=" in param:
                log.error(
                    'missing value for parameter "{}" - ignoring this parameter'.format(
                        param
                    )
                )
                continue
            key, value = param.split("=", 1)
            self.set(key, value)

    """Loads parameters from an init-style configuration file

    :param filename: path to the file to load
    """

    def load_config(self, filename):
        if os.path.exists(filename):
            log.info("loading {}".format(filename))
            tmp = RawConfigParser()
            tmp.read(u"{}".format(filename))

            if tmp.has_section("module-parameters"):
                for key, value in tmp.items("module-parameters"):
                    self.set(key, value)

    """Returns a list of configured modules

    :return: list of configured (active) modules
    :rtype: list of strings
    """

    def modules(self):
        return [item for sub in self.__args.modules for item in sub]

    """Returns the global update interval

    :return: update interval in seconds
    :rtype: float
    """

    def interval(self, default=1):
        return util.format.seconds(self.get("interval", default))

    """Returns whether debug mode is enabled

    :return: True if debug is enabled, False otherwise
    :rtype: boolean
    """

    def debug(self):
        return self.__args.debug

    """Returns whether module order should be reversed/inverted

    :return: True if modules should be reversed, False otherwise
    :rtype: boolean
    """

    def reverse(self):
        return self.__args.right_to_left

    """Returns the logfile location

    :return: location where the logfile should be written
    :rtype: string
    """

    def logfile(self):
        return self.__args.logfile

    """Returns the configured theme name

    :return: name of the configured theme
    :rtype: string
    """

    def theme(self):
        return self.__args.theme

    """Returns the configured iconset name

    :return: name of the configured iconset
    :rtype: string
    """

    def iconset(self):
        return self.__args.iconset

    """Returns which modules should be hidden if their state is not warning/critical

    :return: list of modules to hide automatically
    :rtype: list of strings
    """

    def autohide(self, name):
        return name in self.__args.autohide


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays DNF package update information (<security>/<bugfixes>/<enhancements>/<other>)

Requires the following executable:
    * dnf

Parameters:
    * dnf.interval: Time in minutes between two consecutive update checks (defaults to 30 minutes)

"""

import threading

import core.event
import core.module
import core.widget
import core.decorators

import util.cli


def get_dnf_info(widget):
    res = util.cli.execute("dnf updateinfo", ignore_errors=True)

    security = 0
    bugfixes = 0
    enhancements = 0
    other = 0
    for line in res.split("\n"):
        if not line.startswith(" "):
            continue
        elif "ecurity" in line:
            for s in line.split():
                if s.isdigit():
                    security += int(s)
        elif "ugfix" in line:
            for s in line.split():
                if s.isdigit():
                    bugfixes += int(s)
        elif "hancement" in line:
            for s in line.split():
                if s.isdigit():
                    enhancements += int(s)
        else:
            for s in line.split():
                if s.isdigit():
                    other += int(s)

    widget.set("security", security)
    widget.set("bugfixes", bugfixes)
    widget.set("enhancements", enhancements)
    widget.set("other", other)

    core.event.trigger("update", [widget.module.id], redraw_only=True)


class Module(core.module.Module):
    @core.decorators.every(minutes=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.updates))

    def updates(self, widget):
        result = []
        for t in ["security", "bugfixes", "enhancements", "other"]:
            result.append(str(widget.get(t, 0)))
        return "/".join(result)

    def update(self):
        thread = threading.Thread(target=get_dnf_info, args=(self.widget(),))
        thread.start()

    def state(self, widget):
        cnt = 0
        for t in ["security", "bugfixes", "enhancements", "other"]:
            cnt += widget.get(t, 0)
        if cnt == 0:
            return "good"
        if cnt > 50 or widget.get("security", 0) > 0:
            return "critical"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903,W1401

r""" Execute command in shell and print result

Few command examples:
    'ping -c 1 1.1.1.1 | grep -Po '(?<=time=)\d+(\.\d+)? ms''
    'echo 'BTC=$(curl -s rate.sx/1BTC | grep -Po \'^\d+\')USD''
    'curl -s https://wttr.in/London?format=%l+%t+%h+%w'
    'pip3 freeze | wc -l'
    'any_custom_script.sh | grep arguments'

Parameters:
    * shell.command:  Command to execute
      Use single parentheses if evaluating anything inside (sh-style)
      For example shell.command='echo $(date +'%H:%M:%S')'
      But NOT shell.command='echo $(date +'%H:%M:%S')'
      Second one will be evaluated only once at startup
    * shell.interval: Update interval in seconds
      (defaults to 1s == every bumblebee-status update)
    * shell.async:    Run update in async mode. Won't run next thread if
      previous one didn't finished yet. Useful for long
      running scripts to avoid bumblebee-status freezes
      (defaults to False)

contributed by `rrhuffy <https://github.com/rrhuffy>`_ - many thanks!
"""

import os
import subprocess
import threading

import core.module
import core.widget
import core.input
import util.format
import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.get_output))

        self.__command = self.parameter("command", 'echo "no command configured"')
        self.__async = util.format.asbool(self.parameter("async"))

        if self.__async:
            self.__output = "please wait..."
            self.__current_thread = threading.Thread()

        # LMB and RMB will update output regardless of timer
        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.update)
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.update)

    def set_output(self, value):
        self.__output = value

    def get_output(self, _):
        return self.__output

    def update(self):
        # if requested then run not async version and just execute command in this thread
        if not self.__async:
            self.__output = util.cli.execute(self.__command, shell=True, ignore_errors=True).strip()
            return

        # if previous thread didn't end yet then don't do anything
        if self.__current_thread.is_alive():
            return

        # spawn new thread to execute command and pass callback method to get output from it
        self.__current_thread = threading.Thread(
            target=lambda obj, cmd: obj.set_output(
                util.cli.execute(cmd, ignore_errors=True)
            ),
            args=(self, self.__command),
        )
        self.__current_thread.start()

    def state(self, _):
        if self.__output == "no command configured":
            return "warning"


# vim: tabstop=4 expandtab shiftwidth=4 softtabstop=4



# -*- coding: UTF-8 -*-
# pylint: disable=C0111,R0903

"""Display a stock quote from worldtradingdata.com

Requires the following python packages:
    * requests

Parameters:
    * stock.symbols : Comma-separated list of symbols to fetch
    * stock.change : Should we fetch change in stock value (defaults to True)


contributed by `msoulier <https://github.com/msoulier>`_ - many thanks!
"""

import json
import urllib.request

import logging

import core.module
import core.widget
import core.decorators

import util.format


class Module(core.module.Module):
    @core.decorators.every(hours=1)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.value))

        self.__symbols = self.parameter("symbols", "")
        self.__change = util.format.asbool(self.parameter("change", True))
        self.__value = None

    def value(self, widget):
        results = []
        if not self.__value:
            return "n/a"
        data = json.loads(self.__value)

        for symbol in data["quoteResponse"]["result"]:
            valkey = "regularMarketChange" if self.__change else "regularMarketPrice"
            sym = symbol.get("symbol", "n/a")
            currency = symbol.get("currency", "USD")
            val = "n/a" if not valkey in symbol else "{:.2f}".format(symbol[valkey])
            results.append("{} {} {}".format(sym, val, currency))
        return " ".join(results)

    def fetch(self):
        if self.__symbols:
            url = "https://query1.finance.yahoo.com/v7/finance/quote?symbols="
            url += (
                self.__symbols
                + "&fields=regularMarketPrice,currency,regularMarketChange"
            )
            return urllib.request.urlopen(url).read().strip()
        else:
            logging.error("unable to retrieve stock exchange rate")
            return None

    def update(self):
        self.__value = self.fetch()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Multiwidget CPU module

Can display any combination of:

    * max CPU frequency
    * total CPU load in percents (integer value)
    * per-core CPU load as graph - either mono or colored
    * CPU temperature (in Celsius degrees)
    * CPU fan speed

Requirements:

    * the psutil Python module for the first three items from the list above
    * sensors executable for the rest

Parameters:
    * cpu2.layout: Space-separated list of widgets to add.
      Possible widgets are:

         * cpu2.maxfreq
         * cpu2.cpuload
         * cpu2.coresload
         * cpu2.temp
         * cpu2.fanspeed
    * cpu2.colored: 1 for colored per core load graph, 0 for mono (default)
    * cpu2.temp_pattern: pattern to look for in the output of 'sensors -u';
      required if cpu2.temp widged is used
    * cpu2.fan_pattern: pattern to look for in the output of 'sensors -u';
      required if cpu2.fanspeed widged is used

Note: if you are getting 'n/a' for CPU temperature / fan speed, then you're
lacking the aforementioned pattern settings or they have wrong values.

contributed by `somospocos <https://github.com/somospocos>`_ - many thanks!
"""

import psutil

import core.module

import util.cli
import util.graph
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__layout = self.parameter(
            "layout", "cpu2.maxfreq cpu2.cpuload cpu2.coresload cpu2.temp cpu2.fanspeed"
        )
        self.__widget_names = self.__layout.split()
        self.__colored = util.format.asbool(self.parameter("colored", False))
        for widget_name in self.__widget_names:
            if widget_name == "cpu2.maxfreq":
                widget = self.add_widget(name=widget_name, full_text=self.maxfreq)
                widget.set("type", "freq")
            elif widget_name == "cpu2.cpuload":
                widget = self.add_widget(name=widget_name, full_text=self.cpuload)
                widget.set("type", "load")
            elif widget_name == "cpu2.coresload":
                widget = self.add_widget(name=widget_name, full_text=self.coresload)
                widget.set("type", "loads")
            elif widget_name == "cpu2.temp":
                widget = self.add_widget(name=widget_name, full_text=self.temp)
                widget.set("type", "temp")
            elif widget_name == "cpu2.fanspeed":
                widget = self.add_widget(name=widget_name, full_text=self.fanspeed)
                widget.set("type", "fan")
            if self.__colored:
                widget.set("pango", True)
        self.__temp_pattern = self.parameter("temp_pattern")
        if self.__temp_pattern is None:
            self.__temp = "n/a"
        self.__fan_pattern = self.parameter("fan_pattern")
        if self.__fan_pattern is None:
            self.__fan = "n/a"
        # maxfreq is loaded only once at startup
        if "cpu2.maxfreq" in self.__widget_names:
            self.__maxfreq = psutil.cpu_freq().max / 1000

    def maxfreq(self, _):
        return "{:.2f}GHz".format(self.__maxfreq)

    def cpuload(self, _):
        return "{:>3}%".format(self.__cpuload)

    def add_color(self, bar):
        """add color as pango markup to a bar"""
        if bar in ["▁", "▂"]:
            color = self.theme.color("green", "green")
        elif bar in ["▃", "▄"]:
            color = self.theme.color("yellow", "yellow")
        elif bar in ["▅", "▆"]:
            color = self.theme.color("orange", "orange")
        elif bar in ["▇", "█"]:
            color = self.theme.color("red", "red")
        colored_bar = '<span foreground="{}">{}</span>'.format(color, bar)
        return colored_bar

    def coresload(self, _):
        mono_bars = [util.graph.hbar(x) for x in self.__coresload]
        if not self.__colored:
            return "".join(mono_bars)
        colored_bars = [self.add_color(x) for x in mono_bars]
        return "".join(colored_bars)

    def temp(self, _):
        if self.__temp == "n/a" or self.__temp == 0:
            return "n/a"
        return "{}°C".format(self.__temp)

    def fanspeed(self, _):
        if self.__fanspeed == "n/a":
            return "n/a"
        return "{}RPM".format(self.__fanspeed)

    def _parse_sensors_output(self):
        output = util.cli.execute("sensors -u")
        lines = output.split("\n")
        temp = "n/a"
        fan = "n/a"
        temp_line = None
        fan_line = None
        for line in lines:
            if self.__temp_pattern is not None and self.__temp_pattern in line:
                temp_line = line
            if self.__fan_pattern is not None and self.__fan_pattern in line:
                fan_line = line
            if temp_line is not None and fan_line is not None:
                break
        if temp_line is not None:
            temp = round(float(temp_line.split(":")[1].strip()))
        if fan_line is not None:
            fan = int(fan_line.split(":")[1].strip()[:-4])
        return temp, fan

    def update(self):
        if "cpu2.maxfreq" in self.__widget_names:
            self.__maxfreq = psutil.cpu_freq().max / 1000
        if "cpu2.cpuload" in self.__widget_names:
            self.__cpuload = round(psutil.cpu_percent(percpu=False))
        if "cpu2.coresload" in self.__widget_names:
            self.__coresload = psutil.cpu_percent(percpu=True)
        if "cpu2.temp" in self.__widget_names or "cpu2.fanspeed" in self.__widget_names:
            self.__temp, self.__fanspeed = self._parse_sensors_output()

    def state(self, widget):
        """for having per-widget icons"""
        return [widget.get("type", "")]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays information about the current song in cmus.

Requires the following executable:
    * cmus-remote

Parameters:
    * cmus.format: Format string for the song information. Tag values can be put in curly brackets (i.e. {artist})

      Additional tags:
        * {file} - full song file name
        * {file1} - song file name without path prefix
          if {file} = '/foo/bar.baz', then {file1} = 'bar.baz'
        * {file2} - song file name without path prefix and extension suffix
          if {file} = '/foo/bar.baz', then {file2} = 'bar'
    * cmus.layout: Space-separated list of widgets to add. Possible widgets are the buttons/toggles cmus.prev, cmus.next, cmus.shuffle and cmus.repeat, and the main display with play/pause function cmus.main.
    * cmus.server: The address of the cmus server, either a UNIX socket or host[:port]. Connects to the local instance by default.
    * cmus.passwd: The password to use for the TCP/IP connection.

contributed by `TheEdgeOfRage <https://github.com/TheEdgeOfRage>`_ - many thanks!
"""

from collections import defaultdict

import os
import string

import core.module
import core.input
import core.decorators

import util.cli
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self._layout = self.parameter(
            "layout", "cmus.prev cmus.main cmus.next cmus.shuffle cmus.repeat"
        )
        self._fmt = self.parameter("format", "{artist} - {title} {position}/{duration}")
        self._server = self.parameter("server", None)
        self._passwd = self.parameter("passwd", None)
        self._status = None
        self._shuffle = False
        self._repeat = False
        self._tags = defaultdict(lambda: "")

        # Create widgets
        widget_map = {}
        for widget_name in self._layout.split():
            widget = self.add_widget(name=widget_name)
            self._cmd = "cmus-remote"
            if self._server is not None:
                self._cmd = "{cmd} --server {server}".format(
                    cmd=self._cmd, server=self._server
                )
                if self._passwd is not None:
                    self._cmd = "{cmd} --passwd {passwd}".format(
                        cmd=self._cmd, passwd=self._passwd
                    )

            if widget_name == "cmus.prev":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "{cmd} -r".format(cmd=self._cmd),
                }
            elif widget_name == "cmus.main":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "{cmd} -u".format(cmd=self._cmd),
                }
                widget.full_text(self.description)
            elif widget_name == "cmus.next":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "{cmd} -n".format(cmd=self._cmd),
                }
            elif widget_name == "cmus.shuffle":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "{cmd} -S".format(cmd=self._cmd),
                }
            elif widget_name == "cmus.repeat":
                widget_map[widget] = {
                    "button": core.input.LEFT_MOUSE,
                    "cmd": "{cmd} -R".format(cmd=self._cmd),
                }
            else:
                raise KeyError(
                    "The cmus module does not support a {widget_name!r} widget".format(
                        widget_name=widget_name
                    )
                )

        # Register input callbacks
        for widget, callback_options in widget_map.items():
            core.input.register(widget, **callback_options)

    def hidden(self):
        return self._status is None

    @core.decorators.scrollable
    def description(self, widget):
        return string.Formatter().vformat(self._fmt, (), self._tags)

    def update(self):
        self._load_song()

    def state(self, widget):
        returns = {
            "cmus.shuffle": "shuffle-on" if self._shuffle else "shuffle-off",
            "cmus.repeat": "repeat-on" if self._repeat else "repeat-off",
            "cmus.prev": "prev",
            "cmus.next": "next",
        }
        return returns.get(widget.name, self._status)

    def _eval_line(self, line):
        if line.startswith("file "):
            full_file = line[5:]
            file1 = os.path.basename(full_file)
            file2 = os.path.splitext(file1)[0]
            self._tags.update({"file": full_file})
            self._tags.update({"file1": file1})
            self._tags.update({"file2": file2})
            return
        name, key, value = (line.split(" ", 2) + [None, None])[:3]

        if name == "status":
            self._status = key
        if name == "tag":
            self._tags.update({key: value})
        if name in ["duration", "position"]:
            self._tags.update({name: util.format.duration(int(key))})
        if name == "set" and key == "repeat":
            self._repeat = value == "true"
        if name == "set" and key == "shuffle":
            self._shuffle = value == "true"

    def _load_song(self):
        info = ""
        try:
            info = util.cli.execute("{cmd} -Q".format(cmd=self._cmd))
        except RuntimeError:
            self._status = None

        self._tags = defaultdict(lambda: "")
        for line in info.split("\n"):
            self._eval_line(line)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.input


@pytest.fixture
def obj():
    return core.input.Object()


@pytest.fixture
def obj2():
    return core.input.Object()


@pytest.fixture
def cb(mocker):
    return mocker.MagicMock()


@pytest.fixture
def cb2(mocker):
    return mocker.MagicMock()


def event(input_object):
    return {"button": core.input.LEFT_MOUSE, "instance": input_object.id}


def event2(input_object):
    return {"button": core.input.RIGHT_MOUSE, "instance": input_object.id}


def test_callable_gets_called(obj, cb):
    core.input.register(obj, event(obj)["button"], cb)
    core.input.trigger(event(obj))

    cb.assert_called_once_with(event(obj))


def test_nonexistent_callback(obj, obj2, cb):
    core.input.register(obj, event(obj)["button"], cb)
    core.input.trigger(event(obj2))

    cb.assert_not_called()


def test_different_events(obj, obj2, cb, cb2):
    core.input.register(obj, event(obj)["button"], cb)
    core.input.register(obj, event2(obj)["button"], cb)
    core.input.register(obj2, event(obj)["button"], cb2)
    core.input.register(obj2, event2(obj)["button"], cb2)

    core.input.trigger(event(obj))
    core.input.trigger(event2(obj))

    cb.assert_any_call(event(obj))
    cb.assert_any_call(event2(obj))
    cb2.assert_not_called()


def test_multiple_registrations(obj, cb, cb2):
    core.input.register(obj, event(obj)["button"], cb)
    core.input.register(obj, event(obj)["button"], cb2)

    core.input.trigger(event(obj))

    cb.assert_called_once_with(event(obj))
    cb2.assert_called_once_with(event(obj))


def test_event_names():
    assert core.input.button_name(core.input.LEFT_MOUSE) == "left-mouse"
    assert core.input.button_name(core.input.RIGHT_MOUSE) == "right-mouse"
    assert core.input.button_name(core.input.MIDDLE_MOUSE) == "middle-mouse"
    assert core.input.button_name(core.input.WHEEL_UP) == "wheel-up"
    assert core.input.button_name(core.input.WHEEL_DOWN) == "wheel-down"
    assert core.input.button_name(12345) == "n/a"


def test_non_callable_callback(mocker, obj):
    cli = mocker.patch("core.input.util.cli")
    cli.execute.return_value = ""

    core.input.register(obj, event(obj)["button"], "sample-command")

    core.input.trigger(event(obj))

    cli.execute.assert_called_once_with("sample-command", wait=False, shell=True)


def test_non_existent_callback(mocker, obj):
    cli = mocker.patch("core.input.util.cli")
    cli.execute.return_value = ""
    cli.execute.side_effect = RuntimeError("some-error")

    core.input.register(obj, event(obj)["button"], "sample-command")

    core.input.trigger(event(obj))

    cli.execute.assert_called_once_with("sample-command", wait=False, shell=True)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.config
import modules.contrib.mpd


@pytest.fixture
def mpd_module():
    return modules.contrib.mpd.Module(config=core.config.Config([]), theme=None)


def _test_state(mocker, mpd_module, widget_name, expected_output):
    widget = mocker.Mock()
    widget.name = widget_name
    return mpd_module.state(widget) == expected_output


def test_states(mocker, mpd_module):
    assert _test_state(mocker, mpd_module, "mpd.repeat", "repeat-off")
    assert _test_state(mocker, mpd_module, "mpd.shuffle", "shuffle-off")
    assert _test_state(mocker, mpd_module, "mpd.prev", "prev")
    assert _test_state(mocker, mpd_module, "mpd.next", "next")


def test_no_host(mpd_module):
    assert mpd_module._hostcmd == ""


def test_host():
    module_with_host = modules.contrib.mpd.Module(
        config=core.config.Config(["-p", "mpd.host=sample-host"]), theme=None
    )
    assert module_with_host._hostcmd == " -h sample-host"


def test_host2(mocker):
    cli = mocker.patch("modules.contrib.mpd.util.cli")
    module_with_host = modules.contrib.mpd.Module(
        config=core.config.Config(["-p", "mpd.host=sample-host"]), theme=None
    )
    module_with_host.update()
    args, kwargs = cli.execute.call_args

    assert " -h sample-host" in args[0] and "mpc" in args[0]


def test_bad_layout():
    pytest.raises(
        KeyError,
        modules.contrib.mpd.Module,
        config=core.config.Config(["-p", 'mpd.layout="mpd.inexistent"']),
        theme=None,
    )


def test_hidden_on_creation(mpd_module):
    assert mpd_module.hidden()


def test_update_calls_load_song(mocker, mpd_module):
    mocker.patch.object(mpd_module, "_load_song")
    mpd_module.update()
    mpd_module._load_song.assert_called_with()


def test_default_layout(mpd_module):
    assert mpd_module._layout == "mpd.prev mpd.main mpd.next mpd.shuffle mpd.repeat"



"""get volume level or control it

Parameters:
    * amixer.device: Device to use (default is Master,0)
    * amixer.percent_change: How much to change volume by when scrolling on the module (default is 4%)

contributed by `zetxx <https://github.com/zetxx>`_ - many thanks!

input handling contributed by `ardadem <https://github.com/ardadem>`_ - many thanks!
"""
import re

import core.module
import core.widget
import core.input

import util.cli
import util.format

class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.volume))

        self.__level = "n/a"
        self.__muted = True
        self.__device = self.parameter("device", "Master,0")
        self.__change = util.format.asint(
            self.parameter("percent_change", "4%").strip("%"), 0, 100
        )

        events = [
            {
                "type": "mute",
                "action": self.toggle,
                "button": core.input.LEFT_MOUSE,
            },
            {
                "type": "volume",
                "action": self.increase_volume,
                "button": core.input.WHEEL_UP,
            },
            {
                "type": "volume",
                "action": self.decrease_volume,
                "button": core.input.WHEEL_DOWN,
            },
        ]

        for event in events:
            core.input.register(self, button=event["button"], cmd=event["action"])

    def toggle(self, event):
        self.set_parameter("toggle")

    def increase_volume(self, event):
        self.set_parameter("{}%+".format(self.__change))

    def decrease_volume(self, event):
        self.set_parameter("{}%-".format(self.__change))

    def set_parameter(self, parameter):
        util.cli.execute("amixer -q set {} {}".format(self.__device, parameter))

    def volume(self, widget):
        if self.__level == "n/a":
            return self.__level
        m = re.search(r"([\d]+)\%", self.__level)
        self.__muted = True
        if m:
            if m.group(1) != "0" and "[on]" in self.__level:
                self.__muted = False
            return "{}%".format(m.group(1))
        else:
            return "0%"

    def update(self):
        try:
            self.__level = util.cli.execute(
                "amixer get {}".format(self.__device)
            )
        except Exception as e:
            self.__level = "n/a"

    def state(self, widget):
        if self.__muted:
            return ["warning", "muted"]
        return ["unmuted"]

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



__callbacks = {}


def register(event, callback, *args, **kwargs):
    cb = callback
    if args or kwargs:
        cb = lambda: callback(*args, **kwargs)

    __callbacks.setdefault(event, []).append(cb)


def clear():
    __callbacks.clear()


def trigger(event, *args, **kwargs):
    cb = __callbacks.get(event, [])
    if len(cb) == 0:
        return False

    for callback in cb:
        if args or kwargs:
            callback(*args, **kwargs)
        else:
            callback()
    return True


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Opens a random xkcd comic in the browser.

contributed by `whzup <https://github.com/whzup>`_ - many thanks!
"""

import core.module
import core.widget
import core.input
import core.decorators


class Module(core.module.Module):
    @core.decorators.never
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget("xkcd"))
        core.input.register(
            self,
            button=core.input.LEFT_MOUSE,
            cmd="xdg-open https://c.xkcd.com/random/comic/",
        )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Store interface

Allows arbitrary classes to offer a simple get/set
store interface by deriving from the Store class in
this module
"""


class Store(object):
    """Interface for storing and retrieving simple values"""

    def __init__(self):
        super(Store, self).__init__()
        self._data = {}

    def set(self, key, value):
        """Sets key to value, overwriting any existing data for that key

        :param key: the name of the parameter to set
        :param value: the value to be set
        """
        self._data[key] = {"value": value, "used": False}

    def unused_keys(self):
        """Returns a list of unused keys

        :return: a list of keys that are set, but never used
        :rtype: list of strings
        """
        return [key for key, value in self._data.items() if value["used"] == False]

    def get(self, key, default=None):
        """Returns the current value for the specified key, or a default value,
        if the key is not set

        :param key: the name of the parameter to retrieve
        :param default: the default value to return, defaults to None
        """
        if key in self._data:
            self._data[key]["used"] = True
        return self._data.get(key, {"value": default})["value"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.decorators
import core.widget
import core.module
import core.config


@pytest.fixture
def module():
    class TestModule(core.module.Module):
        @core.decorators.never
        def __init__(self, config=None, theme=None):
            config = core.config.Config([])
            super().__init__(config, theme, core.widget.Widget(self.get))
            self.text = ""

        @core.decorators.scrollable
        def get(self, widget):
            return self.text

    module = TestModule()
    module.set("scrolling.width", 10)
    return module


def test_never(module):
    assert module.parameter("interval") == "never"


def test_no_text(module):
    assert module.text == ""
    assert module.get(module.widget()) == ""


def test_smaller(module):
    module.text = "test"
    assert module.parameter("scrolling.width") > len(module.text)
    assert module.get(module.widget()) == module.text


def test_bigger(module):
    module.text = "this is a really really long sample text"
    maxwidth = module.parameter("scrolling.width")
    assert maxwidth < len(module.text)
    assert module.get(module.widget()) == module.text[:maxwidth]


def test_bounce(module):
    module.text = "abcd"
    module.set("scrolling.width", 2)
    assert module.get(module.widget()) == "ab"
    assert module.get(module.widget()) == "bc"
    assert module.get(module.widget()) == "cd"
    assert module.get(module.widget()) == "bc"
    assert module.get(module.widget()) == "ab"
    assert module.get(module.widget()) == "bc"
    assert module.get(module.widget()) == "cd"
    assert module.get(module.widget()) == "bc"
    assert module.get(module.widget()) == "ab"


def test_nobounce(module):
    module.set("scrolling.bounce", False)
    module.set("scrolling.width", 2)
    module.text = "abcd"

    assert module.get(module.widget()) == "ab"
    assert module.get(module.widget()) == "bc"
    assert module.get(module.widget()) == "cd"
    assert module.get(module.widget()) == "ab"
    assert module.get(module.widget()) == "bc"
    assert module.get(module.widget()) == "cd"


def test_completely_changed_data(module):
    module.text = "abcd"
    module.set("scrolling.width", 2)

    assert module.get(module.widget()) == "ab"
    assert module.get(module.widget()) == "bc"

    module.text = "wxyz"
    assert module.get(module.widget()) == "wx"
    assert module.get(module.widget()) == "xy"


def test_slightly_changed_data(module):
    module.text = "this is a sample song (0:00)"
    module.set("scrolling.width", 10)

    assert module.get(module.widget()) == module.text[0:10]
    module.text = "this is a sample song (0:01)"
    assert module.get(module.widget()) == module.text[1:11]
    module.text = "this is a sample song (0:02)"
    assert module.get(module.widget()) == module.text[2:12]
    module.text = "this is a sample song (0:13)"
    assert module.get(module.widget()) == module.text[3:13]
    module.text = "this is a different song (0:13)"
    assert module.get(module.widget()) == module.text[0:10]


def test_n_plus_one(module):
    module.text = "10 letters"
    module.set("scrolling.width", 9)

    assert module.get(module.widget()) == module.text[0:9]
    assert module.get(module.widget()) == module.text[1:10]
    assert module.get(module.widget()) == module.text[0:9]
    assert module.get(module.widget()) == module.text[1:10]
    assert module.get(module.widget()) == module.text[0:9]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays volume and mute status and controls for PulseAudio devices. Use wheel up and down to change volume, left click mutes, right click opens pavucontrol.

Aliases: pasink (use this to control output instead of input), pasource

Parameters:
    * pulseaudio.autostart: If set to 'true' (default is 'false'), automatically starts the pulseaudio daemon if it is not running
    * pulseaudio.percent_change: How much to change volume by when scrolling on the module (default is 2%)
    * pulseaudio.limit: Upper limit for setting the volume (default is 0%, which means 'no limit')
      Note: If the left and right channels have different volumes, the limit might not be reached exactly.
    * pulseaudio.showbars: 1 for showing volume bars, requires --markup=pango;
      0 for not showing volume bars (default)

Requires the following executable:
    * pulseaudio
    * pactl
    * pavucontrol
"""

import re
import logging

import core.module
import core.widget
import core.input

import util.cli
import util.graph
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme, channel):
        super().__init__(config, theme, core.widget.Widget(self.volume))

        if util.format.asbool(self.parameter("autostart", False)):
            util.cli.execute("pulseaudio --start", ignore_errors=True)

        self._change = util.format.asint(
            self.parameter("percent_change", "2%").strip("%"), 0, 100
        )
        self._limit = util.format.asint(self.parameter("limit", "0%").strip("%"), 0)

        self._left = 0
        self._right = 0
        self._mono = 0
        self._mute = False
        self._failed = False
        self._channel = channel
        self._showbars = util.format.asbool(self.parameter("showbars", 0))

        self._patterns = [
            {"expr": "Name:", "callback": (lambda line: False)},
            {
                "expr": "Mute:",
                "callback": (
                    lambda line: self.mute(False if " no" in line.lower() else True)
                ),
            },
            {"expr": "Volume:", "callback": self.getvolume},
        ]

        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd="pavucontrol")

        events = [
            {"type": "mute", "action": self.toggle, "button": core.input.LEFT_MOUSE},
            {
                "type": "volume",
                "action": self.increase_volume,
                "button": core.input.WHEEL_UP,
            },
            {
                "type": "volume",
                "action": self.decrease_volume,
                "button": core.input.WHEEL_DOWN,
            },
        ]

        for event in events:
            core.input.register(self, button=event["button"], cmd=event["action"])

    def set_volume(self, amount):
        util.cli.execute(
            "pactl set-{}-{} @DEFAULT_{}@ {}".format(
                self._channel, "volume", self._channel.upper(), amount
            )
        )

    def increase_volume(self, event):
        if self._limit > 0:  # we need to check the limit
            left = int(self._left)
            right = int(self._right)
            if (
                left + self._change >= self._limit
                or right + self._change >= self._limit
            ):
                if left == right:
                    # easy case, just set to limit
                    self.set_volume("{}%".format(self._limit))
                    return
                else:
                    # don't adjust anymore, since i don't know how to update only one channel
                    return

        self.set_volume("+{}%".format(self._change))

    def decrease_volume(self, event):
        self.set_volume("-{}%".format(self._change))

    def toggle(self, event):
        util.cli.execute(
            "pactl set-{}-mute @DEFAULT_{}@ toggle".format(
                self._channel, self._channel.upper()
            )
        )

    def mute(self, value):
        self._mute = value

    def getvolume(self, line):
        if "mono" in line:
            m = re.search(r"mono:.*\s*\/\s*(\d+)%", line)
            if m:
                self._mono = m.group(1)
        else:
            m = re.search(r"left:.*\s*\/\s*(\d+)%.*right:.*\s*\/\s*(\d+)%", line)
            if m:
                self._left = m.group(1)
                self._right = m.group(2)

    def _default_device(self):
        output = util.cli.execute("pactl info")
        pattern = "Default {}: ".format("Sink" if self._channel == "sink" else "Source")
        for line in output.split("\n"):
            if line.startswith(pattern):
                return line.replace(pattern, "")
        logging.error("no pulseaudio device found")
        return "n/a"

    def volume(self, widget):
        if self._failed == True:
            return "n/a"
        if int(self._mono) > 0:
            vol = "{}%".format(self._mono)
            if self._showbars:
                vol = "{} {}".format(vol, util.graph.hbar(float(self._mono)))
            return vol
        elif self._left == self._right:
            vol = "{}%".format(self._left)
            if self._showbars:
                vol = "{} {}".format(vol, util.graph.hbar(float(self._left)))
            return vol
        else:
            vol = "{}%/{}%".format(self._left, self._right)
            if self._showbars:
                vol = "{} {}{}".format(
                    vol,
                    util.graph.hbar(float(self._left)),
                    util.graph.hbar(float(self._right)),
                )
            return vol

    def update(self):
        try:
            self._failed = False
            channel = "sinks" if self._channel == "sink" else "sources"
            device = self._default_device()

            result = util.cli.execute("pactl list {}".format(channel))
            found = False

            for line in result.split("\n"):
                if "Name: {}".format(device) in line:
                    found = True
                    continue
                if found is False:
                    continue
                for pattern in self._patterns:
                    if not pattern["expr"] in line:
                        continue
                    if pattern["callback"](line) is False and found == True:
                        return
        except Exception as e:
            self._failed = True
            logging.exception(e)
            if util.format.asbool(self.parameter("autostart", False)):
                util.cli.execute("pulseaudio --start", ignore_errors=True)
            else:
                raise e

    def state(self, widget):
        if self._mute:
            return ["warning", "muted"]
        if int(self._left) > int(100):
            return ["critical", "unmuted"]
        return ["unmuted"]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# -*- coding: UTF-8 -*-
# pylint: disable=C0111,R0903

"""Displays the temperature on the current location based on the ip

Requires the following python packages:
    * requests

Parameters:
    * weather.location: Set location, defaults to 'auto' for getting location automatically from a web service
      If set to a comma-separated list, left-click and right-click can be used to rotate the locations.
      Locations should be city names or city ids.
    * weather.unit: metric (default), kelvin, imperial
    * weather.showcity: If set to true, show location information, otherwise hide it (defaults to true)
    * weather.showminmax: If set to true, show the minimum and maximum temperature, otherwise hide it (defaults to false)
    * weather.apikey: API key from http://api.openweathermap.org


contributed by `TheEdgeOfRage <https://github.com/TheEdgeOfRage>`_ - many thanks!
"""

import core.module
import core.widget
import core.input

import util.format
import util.location

import re

import requests
from requests.exceptions import RequestException


class Module(core.module.Module):
    @core.decorators.every(minutes=15)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.output))

        self.__temperature = 0
        self.__apikey = self.parameter("apikey", "af7bfe22287c652d032a3064ffa44088")
        self.__location = util.format.aslist(self.parameter("location", "auto"))

        self.__index = 0
        self.__showcity = util.format.asbool(self.parameter("showcity", True))
        self.__showminmax = util.format.asbool(self.parameter("showminmax", False))
        self.__unit = self.parameter("unit", "metric")
        self.__valid = False

        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd=self.__next_location
        )
        core.input.register(
            self, button=core.input.RIGHT_MOUSE, cmd=self.__prev_location
        )

    def __next_location(self, event):
        self.__index = (self.__index + 1) % len(self.__location)
        self.update()

    def __prev_location(self, event):
        self.__index = (
            len(self.__location) - 1 if self.__index <= 0 else self.__index - 1
        )
        self.update()

    def temperature(self):
        return util.format.astemperature(self.__temperature, self.__unit)

    def tempmin(self):
        return util.format.astemperature(self.__tempmin, self.__unit)

    def tempmax(self):
        return util.format.astemperature(self.__tempmax, self.__unit)

    def city(self):
        city = re.sub(r"[_-]", " ", self.__city)
        return "{} ".format(city)

    def output(self, widget):
        if not self.__valid:
            return "?"
        if self.__showminmax:
            self.__showcity = False
            return (
                self.city()
                + self.temperature()
                + "  Hi:"
                + self.tempmax()
                + "  Lo:"
                + self.tempmin()
            )
        elif self.__showcity:
            return self.city() + self.temperature()
        else:
            return self.temperature()

    def state(self, widget):
        if self.__valid:
            if "thunderstorm" in self.__weather:
                return ["thunder"]
            elif "drizzle" in self.__weather:
                return ["rain"]
            elif "rain" in self.__weather:
                return ["rain"]
            elif "snow" in self.__weather:
                return ["snow"]
            elif "sleet" in self.__weather:
                return ["sleet"]
            elif "clear" in self.__weather:
                return ["clear"]
            elif "cloud" in self.__weather:
                return ["clouds"]

        return []

    def update(self):
        try:
            weather_url = "http://api.openweathermap.org/data/2.5/weather?appid={}".format(
                self.__apikey
            )
            weather_url = "{}&units={}".format(weather_url, self.__unit)
            if self.__location[self.__index] == "auto":
                coord = util.location.coordinates()
                weather_url = "{url}&lat={lat}&lon={lon}".format(
                    url=weather_url, lat=coord[0], lon=coord[1]
                )
            elif self.__location[self.__index].isdigit():
                weather_url = "{url}&id={id}".format(
                    url=weather_url, id=self.__location[self.__index]
                )
            else:
                weather_url = "{url}&q={city}".format(
                    url=weather_url, city=self.__location[self.__index]
                )
            weather = requests.get(weather_url).json()
            self.__city = weather["name"]
            self.__temperature = int(weather["main"]["temp"])
            self.__tempmin = int(weather["main"]["temp_min"])
            self.__tempmax = int(weather["main"]["temp_max"])
            self.__weather = weather["weather"][0]["main"].lower()
            self.__valid = True
        except Exception:
            self.__valid = False


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Draws a widget with configurable text content.

Parameters:
    * spacer.text: Widget contents (defaults to empty string)
"""

import core.module
import core.widget
import core.decorators


class Module(core.module.Module):
    @core.decorators.every(minutes=60)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.text))
        self.__text = self.parameter("text", "")

    def text(self, _):
        return self.__text


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Shows bumblebee-status errors"""

import platform

import core.module
import core.widget
import core.event


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.full_text))
        self.__error = ""
        self.__state = "critical"

        core.event.register("error", self.__set_error)

    def full_text(self, widgets):
        return self.__error

    def __set_error(self, error="n/a", state="critical"):
        self.__error = error
        self.__state = state

    def state(self, widget):
        if self.__error:
            return [self.__state]
        return []


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays bluetooth status (Bluez). Left mouse click launches manager app,
right click toggles bluetooth. Needs dbus-send to toggle bluetooth state.

Parameters:
    * bluetooth.device : the device to read state from (default is hci0)
    * bluetooth.manager : application to launch on click (blueman-manager)
    * bluetooth.dbus_destination : dbus destination (defaults to org.blueman.Mechanism)
    * bluetooth.dbus_destination_path : dbus destination path (defaults to /)
    * bluetooth.right_click_popup : use popup menu when right-clicked (defaults to True)

contributed by `brunosmmm <https://github.com/brunosmmm>`_ - many thanks!
"""


import os
import re
import logging

import core.module
import core.widget
import core.input

import util.cli
import util.format
import util.popup


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.status))

        device = self.parameter("device", "hci0")
        self.manager = self.parameter("manager", "blueman-manager")
        self._path = "/sys/class/bluetooth/{}".format(device)
        self._status = "Off"

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.manager)

        # determine whether to use pop-up menu or simply toggle the device on/off
        right_click_popup = util.format.asbool(
            self.parameter("right_click_popup", True)
        )

        if right_click_popup:
            core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.popup)
        else:
            core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self._toggle)

    def status(self, widget):
        """Get status."""
        return self._status

    def update(self):
        """Update current state."""
        if not os.path.exists(self._path):
            self._status = "?"
            return

        # search for whichever rfkill directory available
        try:
            dirnames = next(os.walk(self._path))[1]
            for dirname in dirnames:
                m = re.match(r"rfkill[0-9]+", dirname)
                if m is not None:
                    with open(os.path.join(self._path, dirname, "state"), "r") as f:
                        state = int(f.read())
                        if state == 1:
                            self._status = "On"
                        else:
                            self._status = "Off"
                    return

        except IOError:
            self._status = "?"

    def popup(self, widget):
        """Show a popup menu."""
        menu = util.popup.PopupMenu()
        if self._status == "On":
            menu.add_menuitem("Disable Bluetooth")
        elif self._status == "Off":
            menu.add_menuitem("Enable Bluetooth")
        else:
            return

        # show menu and get return code
        ret = menu.show(widget)
        if ret == 0:
            # first (and only) item selected.
            self._toggle()

    def _toggle(self, widget=None):
        """Toggle bluetooth state."""
        if self._status == "On":
            state = "false"
        else:
            state = "true"

        dst = self.parameter("dbus_destination", "org.blueman.Mechanism")
        dst_path = self.parameter("dbus_destination_path", "/")

        cmd = (
            "dbus-send --system --print-reply --dest={}"
            " {} org.blueman.Mechanism.SetRfkillState"
            " boolean:{}".format(dst, dst_path, state)
        )

        logging.debug("bt: toggling bluetooth")
        util.cli.execute(cmd)

    def state(self, widget):
        """Get current state."""
        state = []

        if self._status == "?":
            state = ["unknown"]
        elif self._status == "On":
            state = ["ON"]
        else:
            state = ["OFF"]

        return state


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the brightness of a display

Parameters:
    * brightness.step: The amount of increase/decrease on scroll in % (defaults to 2)
    * brightness.device_path: The device path (defaults to /sys/class/backlight/intel_backlight), can contain wildcards (in this case, the first matching path will be used); This is only used when brightness.use_acpi is set to true
    * brightness.use_acpi: If set to true, read brightness directly from the sys ACPI interface, using the device specified in brightness.device_path (defaults to false)

contributed by `TheEdgeOfRage <https://github.com/TheEdgeOfRage>`_ - many thanks!
"""

import glob
import shutil

import core.module
import core.widget
import core.input
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(seconds=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.brightness))

        self.__brightness = "n/a"
        self.__readcmd = None
        step = self.parameter("step", 2)
        self.__device_path = self.find_device(self.parameter("device_path", "/sys/class/backlight/intel_backlight"))

        if util.format.asbool(self.parameter("use_acpi", False)):
            self.__readcmd = self.__acpi
            # TODO: add setting
        elif shutil.which("light"):
            self.__readcmd = self.__light
            self.register_cmd("light -A {}%".format(step), "light -U {}%".format(step))
        elif shutil.which("brightnessctl"):
            self.__readcmd = self.__brightnessctl
            self.register_cmd(
                "brightnessctl s {}%+".format(step), "brightnessctl s {}%-".format(step)
            )
        else:
            self.__readcmd = self.__xbacklight
            self.register_cmd(
                "xbacklight +{}%".format(step), "xbacklight -{}%".format(step)
            )

    def find_device(self, device_path):
        res = glob.glob(device_path)
        if len(res) == 0:
            return device_path
        return res[0]

    def register_cmd(self, up_cmd, down_cmd):
        core.input.register(self, button=core.input.WHEEL_UP, cmd=up_cmd)
        core.input.register(self, button=core.input.WHEEL_DOWN, cmd=down_cmd)

    def brightness(self, widget):
        return self.__brightness

    def __acpi(self):
        try:
            backlight = 1
            max_brightness = 1
            with open("{}/brightness".format(self.__device_path)) as f:
                backlight = int(f.readline())
            with open("{}/max_brightness".format(self.__device_path)) as f:
                max_brightness = int(f.readline())
                return float(backlight*100)/max_brightness
        except:
            return "unable to read brightness from {}".format(self.__device_path)

    def __light(self):
        return util.cli.execute("light").strip()

    def __brightnessctl(self):
        m = util.cli.execute("brightnessctl m").strip()
        g = util.cli.execute("brightnessctl g").strip()
        return float(g) / float(m) * 100.0

    def __xbacklight(self):
        return util.cli.execute("xbacklight -get").strip()

    def update(self):
        try:
            tmp = self.__readcmd()
            if isinstance(tmp, str):
                self.__brightness = tmp
            else:
                self.__brightness = "{:3.0f}%".format(float(tmp))
        except:
            self.__brightness = "n/a"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903
# -*- coding: utf-8 -*-

"""Displays information about the current song in mocp. Left click toggles play/pause. Right click toggles shuffle.

Requires the following executable:
    * mocp

Parameters:
    * mocp.format: Format string for the song information. Replace string sequences with the actual information:

       * %state     State
       * %file      File
       * %title     Title, includes track, artist, song title and album
       * %artist    Artist
       * %song      SongTitle
       * %album     Album
       * %tt        TotalTime
       * %tl        TimeLeft
       * %ts        TotalSec
       * %ct        CurrentTime
       * %cs        CurrentSec
       * %b         Bitrate
       * %r         Sample rate

contributed by `chrugi <https://github.com/chrugi>`_ - many thanks!
"""

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.description))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd="mocp -G")
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd="mocp -t shuffle")
        self.__format = self.parameter("format", "%state %artist - %song | %ct/%tt")
        self.__running = False

    def description(self, widget):
        return self.__info if self.__running == True else "Music On Console Player"

    def update(self):
        self.__load_song()

    def __load_song(self):
        try:
            self.__info = util.cli.execute("mocp -Q '{}'".format(self.__format)).strip()
            self.__running = True
        except RuntimeError:
            self.__running = False


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import json
import pytest

import core.event
import core.config
import core.output
import core.module


class SampleModule(core.module.Module):
    pass


@pytest.fixture(autouse=True)
def clear_events():
    core.event.clear()


@pytest.fixture
def i3():
    return core.output.i3()


@pytest.fixture
def module_a(mocker):
    widget = mocker.MagicMock()
    widget.full_text.return_value = "test"
    widget.id = "a"
    return SampleModule(config=core.config.Config([]), widgets=[widget, widget, widget])

@pytest.fixture
def module_b(mocker):
    widget = mocker.MagicMock()
    widget.full_text.return_value = "test"
    widget.id = "b"
    return SampleModule(config=core.config.Config([]), widgets=[widget, widget, widget])


@pytest.fixture
def paddedTheme():
    return core.theme.Theme(raw_data={"defaults": {"padding": " "}})


@pytest.fixture
def separatorTheme():
    return core.theme.Theme(
        raw_data={"defaults": {"separator": "***", "fg": "red", "bg": "blue"}}
    )


@pytest.fixture
def block_a(separatorTheme, module_a):
    return core.output.block(
        theme=separatorTheme, module=module_a, widget=module_a.widget(),
    )

def test_start(i3):
    all_data = i3.start()
    data = all_data["blocks"]

    assert data["version"] == 1
    assert data["click_events"] == True
    assert all_data["suffix"] == "\n["


def test_stop(i3):
    assert i3.stop()["suffix"] == "\n]"


def test_no_modules_by_default(i3):
    assert i3.modules() == []


def test_register_single_module(i3, module_a):
    i3.modules(module_a)

    assert i3.modules() == [module_a]


def test_register_multiple_modules(i3, module_a):
    i3.modules([module_a, module_a, module_a])
    assert i3.modules() == [module_a, module_a, module_a]

def test_toggle_module(i3, module_a, module_b):
    i3.modules([module_a, module_b])

    i3.update()
    i3.toggle_minimize({ "instance": module_a.widget().id })
    i3.update()

    assert i3.content()[module_a.widget().id]["minimized"] == True

#    assert module_a.widget().minimized == True
#    assert module_b.widget().minimized == False
#
#    i3.toggle_minimize({ "instance": module_a.widget().id })
#    i3.toggle_minimize({ "instance": module_b.widget().id })
#
#    assert module_a.widget().minimized == False
#    assert module_b.widget().minimized == True

def test_draw_existing_module(mocker, i3):
    i3.test_draw = mocker.MagicMock(
        return_value={"blocks": {"test": True}, "suffix": "end"}
    )
    i3.draw("test_draw")
    i3.test_draw.assert_called_once_with()


def test_empty_status_line(i3):
    data = i3.statusline()

    assert data["blocks"] == []
    assert data["suffix"] == ","


def test_statusline(i3, module_a):
    i3.modules([module_a, module_a, module_a])
    i3.update()
    data = i3.statusline()
    assert len(data["blocks"]) == len(module_a.widgets()) * 3


def test_padding(i3, paddedTheme, module_a):
    i3.theme(paddedTheme)
    blk = core.output.block(i3.theme(), module_a, module_a.widget())
    blk.set("full_text", "abc")
    result = blk.dict()["full_text"]
    assert result == " abc "


def test_no_separator(i3, module_a):
    result = i3.separator_block(module_a, module_a.widget())
    assert result == []


def test_separator(i3, separatorTheme, module_a):
    i3.theme(separatorTheme)
    result = i3.separator_block(module_a, module_a.widget())

    assert len(result) == 1
    assert result[0].dict()["full_text"] == "***"
    assert result[0].dict().get("_decorator") == True
    assert result[0].dict()["color"] == separatorTheme.get("bg", module_a.widget())


def test_dump_json(mocker):
    obj = mocker.MagicMock()
    obj.dict = mocker.MagicMock()
    core.output.dump_json(obj)
    obj.dict_assert_called_once_with()


def test_assign():
    src = {"a": "x", "b": "y", "c": "z"}
    dst = {}

    core.output.assign(src, dst, "a")
    assert src["a"] == dst["a"]

    core.output.assign(src, dst, "123", "b")
    assert src["b"] == dst["123"]

    core.output.assign(src, dst, "blub", default="def")
    assert dst["blub"] == "def"


def test_pango_detection(block_a):
    assert block_a.is_pango({}) == False
    assert block_a.is_pango({"pango": {}}) == True


def test_pangoize(block_a):
    assert block_a.pangoize("test") == "test"
    assert not "markup" in block_a.dict()

    pango = block_a.pangoize({"pango": {"attr": "blub", "x": "y", "full_text": "test"}})
    assert 'attr="blub"' in pango
    assert 'x="y"' in pango
    assert "<span " in pango
    assert ">test</span>" in pango
    assert block_a.dict()["markup"] == "pango"


def test_padding(block_a):
    block_a.set("padding", "***")
    block_a.set("full_text", "test")

    assert block_a.dict()["full_text"] == "***test***"


def test_pre_suffix(block_a):
    block_a.set("padding", "*")
    block_a.set("prefix", "pre")
    block_a.set("suffix", "suf")
    block_a.set("full_text", "test")

    assert block_a.dict()["full_text"] == "*pre*test*suf*"


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Performs a speedtest - only updates when the "play" button is clicked

Requires the following python module:
    * speedtest-cli

"""

import sys

import core.module
import core.widget
import core.input
import core.event
import core.decorators

import speedtest


class Module(core.module.Module):
    @core.decorators.never
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.background = True
        self.__result = "<speedtest>"
        self.__running = False

        start = self.add_widget(name="start")
        main = self.add_widget(name="main", full_text=self.result)

        core.input.register(start, button=core.input.LEFT_MOUSE, cmd=self.update_event)

    def result(self, _):
        return self.__result

    def update_event(self, _):
        self.__running = True
        self.update()

    def update(self):
        if not self.__running:
            return
        core.event.trigger("update", [self.id], redraw_only=True)
        s = speedtest.Speedtest()
        s.get_best_server()
        s.download(threads=None)
        s.upload(threads=None)

        self.__result = "ping: {:.2f}ms down: {:.2f}Mbps up: {:.2f}Mbps".format(
            s.results.ping,
            s.results.download / 1024 / 1024,
            s.results.upload / 1024 / 1024,
        )
        self.__running = False
        core.event.trigger("update", [self.id], redraw_only=True)

    def state(self, widget):
        if widget.name == "start":
            return "running" if self.__running else "not-running"
        return None


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import pytest

import core.config
import modules.contrib.kernel


@pytest.fixture
def some_kernel():
    return "this-is-my-kernel"


@pytest.fixture
def kernel_module():
    return modules.contrib.kernel.Module(config=core.config.Config([]), theme=None)


def test_full_text(mocker, kernel_module):
    platform = mocker.patch("modules.contrib.kernel.platform")
    platform.release.return_value = some_kernel
    assert len(kernel_module.widgets()) == 1
    assert some_kernel == kernel_module.widget().full_text()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays and changes the current keyboard layout

Requires the following executable:
    * setxkbmap

contributed by `Pseudonick47 <https://github.com/Pseudonick47>`_ - many thanks!
"""

import core.module
import core.widget
import core.input

import util.cli


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.current_layout))

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__next_keymap)
        core.input.register(self, button=core.input.RIGHT_MOUSE, cmd=self.__prev_keymap)

    def __next_keymap(self, event):
        self._set_keymap(1)

    def __prev_keymap(self, event):
        self._set_keymap(-1)

    def _set_keymap(self, rotation):
        layouts = self.get_layouts()
        if len(layouts) == 1:
            return  # nothing to do
        layouts = layouts[rotation:] + layouts[:rotation]

        layout_list = []
        variant_list = []
        for l in layouts:
            tmp = l.split(":")
            layout_list.append(tmp[0])
            variant_list.append(tmp[1] if len(tmp) > 1 else "")

        util.cli.execute(
            "setxkbmap -layout {} -variant {}".format(
                ",".join(layout_list), ",".join(variant_list)
            ),
            ignore_errors=True,
        )

    def get_layouts(self):
        try:
            res = util.cli.execute("setxkbmap -query")
        except RuntimeError:
            return ["n/a"]
        layouts = []
        variants = []
        for line in res.split("\n"):
            if not line:
                continue
            if "layout" in line:
                layouts = line.split(":")[1].strip().split(",")
            if "variant" in line:
                variants = line.split(":")[1].strip().split(",")

        result = []
        for idx, layout in enumerate(layouts):
            if len(variants) > idx and variants[idx]:
                layout = "{}:{}".format(layout, variants[idx])
            result.append(layout)
        return result if len(result) > 0 else ["n/a"]

    def current_layout(self, widget):
        layouts = self.get_layouts()
        return layouts[0]


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays update information per repository for pacman.

Parameters:
    * pacman.sum: If you prefere displaying updates with a single digit (defaults to 'False')

Requires the following executables:
    * fakeroot
    * pacman

contributed by `Pseudonick47 <https://github.com/Pseudonick47>`_ - many thanks!
"""

import os
import threading

import core.module
import core.widget
import core.decorators

import util.cli
import util.format

from bumblebee_status.discover import utility

# list of repositories.
# the last one should always be other
repos = ["core", "extra", "community", "multilib", "testing", "other"]


def get_pacman_info(widget, path):
    cmd = utility("pacman-updates")
    result = util.cli.execute(cmd, ignore_errors=True)

    count = len(repos) * [0]

    for line in result.splitlines():
        if line.startswith(("http", "rsync")):
            for i in range(len(repos) - 1):
                if "/" + repos[i] + "/" in line:
                    count[i] += 1
                    break
            else:
                result[-1] += 1

    for i in range(len(repos)):
        widget.set(repos[i], count[i])
    core.event.trigger("update", [widget.module.id], redraw_only=True)


class Module(core.module.Module):
    @core.decorators.every(minutes=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.updates))

    def updates(self, widget):
        if util.format.asbool(self.parameter("sum")):
            return str(sum(map(lambda x: widget.get(x, 0), repos)))
        return "/".join(map(lambda x: str(widget.get(x, 0)), repos))

    def update(self):
        path = os.path.dirname(os.path.abspath(__file__))
        thread = threading.Thread(target=get_pacman_info, args=(self.widget(), path))
        thread.start()

    def state(self, widget):
        weightedCount = sum(
            map(lambda x: (len(repos) - x[0]) * widget.get(x[1], 0), enumerate(repos))
        )

        if weightedCount < 10:
            return "good"

        return self.threshold_state(weightedCount, 100, 150)


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the Octorpint status and the printer's bed/tools temperature in the status bar.

   Left click opens a popup which shows the bed & tools temperatures and additionally a livestream of the webcam (if enabled).

Parameters:
    * octoprint.address     : Octoprint address (e.q: http://192.168.1.3)
    * octoprint.apitoken    : Octorpint API Token (can be obtained from the Octoprint Webinterface)
    * octoprint.webcam      : Set to True if a webcam is connected (default: False)

contributed by `bbernhard <https://github.com/bbernhard>`_ - many thanks!
"""


import urllib
import logging
import threading
import queue

import tkinter as tk
from io import BytesIO
from PIL import Image, ImageTk

import requests
import simplejson

import core.module
import core.widget
import core.input


def get_frame(url):
    img_bytes = b""
    stream = urllib.request.urlopen(url)
    while True:
        img_bytes += stream.read(1024)
        a = img_bytes.find(b"\xff\xd8")
        b = img_bytes.find(b"\xff\xd9")
        if a != -1 and b != -1:
            jpg = img_bytes[a : b + 2]
            img_bytes = img_bytes[b + 2 :]
            img = Image.open(BytesIO(jpg))
            return img
    return None


class WebcamImagesWorker(threading.Thread):
    def __init__(self, url, queue):
        threading.Thread.__init__(self)

        self.__url = url
        self.__queue = queue
        self.__running = True

    def run(self):
        while self.__running:
            img = get_frame(self.__url)
            self.__queue.put(img)

    def stop(self):
        self.__running = False


class Module(core.module.Module):
    @core.decorators.every(seconds=5)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.octoprint_status))

        self.__octoprint_state = "Unknown"
        self.__octoprint_address = self.parameter("address", "")
        self.__octoprint_api_token = self.parameter("apitoken", "")
        self.__octoprint_webcam = self.parameter("webcam", False)

        self.__webcam_images_worker = None
        self.__webcam_image_url = self.__octoprint_address + "/webcam/?action=stream"
        self.__webcam_images_queue = None

        self.__printer_bed_temperature = "-"
        self.__tool1_temperature = "-"

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__show_popup)

    def octoprint_status(self, widget):
        if self.__octoprint_state == "Offline" or self.__octoprint_state == "Unknown":
            return self.__octoprint_state
        return (
            self.__octoprint_state
            + " | B: "
            + str(self.__printer_bed_temperature)
            + "°C"
            + " | T1: "
            + str(self.__tool1_temperature)
            + "°C"
        )

    def __get(self, endpoint):
        url = self.__octoprint_address + "/api/" + endpoint
        headers = {"X-Api-Key": self.__octoprint_api_token}
        resp = requests.get(url, headers=headers)

        try:
            return resp.json(), resp.status_code
        except simplejson.errors.JSONDecodeError:
            return None, resp.status_code

    def __get_printer_bed_temperature(self):
        printer_info, status_code = self.__get("printer")
        if status_code == 200:
            return (
                printer_info["temperature"]["bed"]["actual"],
                printer_info["temperature"]["bed"]["target"],
            )
        return None, None

    def __get_octoprint_state(self):
        job_info, status_code = self.__get("job")
        return job_info["state"] if status_code == 200 else "Unknown"

    def __get_tool_temperatures(self):
        tool_temperatures = []

        printer_info, status_code = self.__get("printer")
        if status_code == 200:
            temperatures = printer_info["temperature"]

            tool_id = 0
            while True:
                try:
                    tool = temperatures["tool" + str(tool_id)]
                    tool_temperatures.append((tool["actual"], tool["target"]))
                except KeyError:
                    break
                tool_id += 1
        return tool_temperatures

    def update(self):
        try:
            self.__octoprint_state = self.__get_octoprint_state()

            actual_temp, _ = self.__get_printer_bed_temperature()
            if actual_temp is None:
                actual_temp = "-"
            self.__printer_bed_temperature = str(actual_temp)

            tool_temps = self.__get_tool_temperatures()
            if len(tool_temps) > 0:
                self.__tool1_temperature = tool_temps[0][0]
            else:
                self.__tool1_temperature = "-"
        except Exception as e:
            logging.exception("Couldn't get data")

    def __refresh_image(self, root, webcam_image, webcam_image_container):
        try:
            img = self.__webcam_images_queue.get()
            webcam_image = ImageTk.PhotoImage(img)
            webcam_image_container.config(image=webcam_image)
        except queue.Empty as e:
            pass
        except Exception as e:
            logging.exception("Couldn't refresh image")

        root.after(5, self.__refresh_image, root, webcam_image, webcam_image_container)

    def __refresh_temperatures(
        self, root, printer_bed_temperature_label, tools_temperature_label
    ):
        actual_bed_temp, target_bed_temp = self.__get_printer_bed_temperature()
        if actual_bed_temp is None:
            actual_bed_temp = "-"
        if target_bed_temp is None:
            target_bed_temp = "-"

        bed_temp = "Bed: " + str(actual_bed_temp) + "/" + str(target_bed_temp) + " °C"
        printer_bed_temperature_label.config(text=bed_temp)

        tool_temperatures = self.__get_tool_temperatures()
        tools_temp = "Tools: "

        if len(tool_temperatures) == 0:
            tools_temp += "-/- °C"
        else:
            for i, tool_temperature in enumerate(tool_temperatures):
                tools_temp += (
                    str(tool_temperature[0]) + "/" + str(tool_temperature[1]) + "°C"
                )
                if i != len(tool_temperatures) - 1:
                    tools_temp += "\t"
        tools_temperature_label.config(text=tools_temp)

        root.after(
            500,
            self.__refresh_temperatures,
            root,
            printer_bed_temperature_label,
            tools_temperature_label,
        )

    def __show_popup(self, widget):
        root = tk.Tk()
        root.attributes("-type", "dialog")
        root.title("Octoprint")
        frame = tk.Frame(root)
        if self.__octoprint_webcam:

            # load first image synchronous before popup is shown, otherwise tkinter isn't able to layout popup properly
            img = get_frame(self.__webcam_image_url)
            webcam_image = ImageTk.PhotoImage(img)
            webcam_image_container = tk.Button(frame, image=webcam_image)
            webcam_image_container.pack()

            self.__webcam_images_queue = queue.Queue()

            self.__webcam_images_worker = WebcamImagesWorker(
                self.__webcam_image_url, self.__webcam_images_queue
            )
            self.__webcam_images_worker.start()
        else:
            logging.debug(
                "Not using webcam, as webcam is disabled. Enable with --webcam."
            )
        frame.pack()

        temperatures_label = tk.Label(frame, text="Temperatures", font=("", 25))
        temperatures_label.pack()

        printer_bed_temperature_label = tk.Label(
            frame, text="Bed: -/- °C", font=("", 15)
        )
        printer_bed_temperature_label.pack()

        tools_temperature_label = tk.Label(frame, text="Tools: -/- °C", font=("", 15))
        tools_temperature_label.pack()

        root.after(10, self.__refresh_image, root, webcam_image, webcam_image_container)
        root.after(
            500,
            self.__refresh_temperatures,
            root,
            printer_bed_temperature_label,
            tools_temperature_label,
        )
        root.bind("<Destroy>", self.__on_close_popup)

        root.eval("tk::PlaceWindow . center")
        root.mainloop()

    def __on_close_popup(self, event):
        self.__webcam_images_queue = None
        self.__webcam_images_worker.stop()


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays information about the current song in vlc, audacious, bmp, xmms2, spotify and others

Requires the following executable:
    * playerctl

contributed by `smitajit <https://github.com/smitajit>`_ - many thanks!

"""

import core.module
import core.widget
import core.input
import util.cli

class Module(core.module.Module):
    def __init__(self,config , theme):
        widgets = [
            core.widget.Widget(name="playerctl.prev"),
            core.widget.Widget(name="playerctl.main", full_text=self.description),
            core.widget.Widget(name="playerctl.next"),
        ]
        super(Module, self).__init__(config, theme ,  widgets)

        core.input.register(widgets[0], button=core.input.LEFT_MOUSE,
            cmd="playerctl previous")
        core.input.register(widgets[1], button=core.input.LEFT_MOUSE,
             cmd="playerctl play-pause")
        core.input.register(widgets[2], button=core.input.LEFT_MOUSE,
             cmd="playerctl next")

        self._status = None
        self._tags = None

    def description(self, widget):
        return self._tags if self._tags else "..."

    def update(self):
        self._load_song()

    def state(self, widget):
        if widget.name == "playerctl.prev":
            return "prev"
        if widget.name == "playerctl.next":
            return "next"
        return self._status

    def _load_song(self):
        info = ""
        try:
            status = util.cli.execute("playerctl status").lower()
            info = util.cli.execute("playerctl metadata xesam:title")
        except :
            self._status = None
            self._tags = None
            return
        self._status = status.split("\n")[0].lower()
        self._tags = info.split("\n")[0][:20]

# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays the price of a cryptocurrency.

Requires the following python packages:
    * requests

Parameters:
    * getcrypto.interval: Interval in seconds for updating the price, default is 120, less than that will probably get your IP banned.
    * getcrypto.getbtc: 0 for not getting price of BTC, 1 for getting it (default).
    * getcrypto.geteth: 0 for not getting price of ETH, 1 for getting it (default).
    * getcrypto.getltc: 0 for not getting price of LTC, 1 for getting it (default).
    * getcrypto.getcur: Set the currency to display the price in, usd is the default.

contributed by `Ryunaq <https://github.com/Ryunaq>`_ - many thanks!
"""

import requests
from requests.exceptions import RequestException
import time

import core.module
import core.widget
import core.input
import core.decorators

import util.format


def getfromkrak(coin, currency):
    abbrev = {
        "Btc": ["xbt", "XXBTZ"],
        "Eth": ["eth", "XETHZ"],
        "Ltc": ["ltc", "XLTCZ"],
    }
    data = abbrev.get(coin, None)
    if not data:
        return
    epair = "{}{}".format(data[0], currency)
    tickname = "{}{}".format(data[1], currency.upper())
    try:
        krakenget = requests.get(
            "https://api.kraken.com/0/public/Ticker?pair=" + epair
        ).json()
    except (RequestException, Exception):
        return "No connection"
    if not "result" in krakenget:
        return "No data"
    kethusdask = float(krakenget["result"][tickname]["a"][0])
    kethusdbid = float(krakenget["result"][tickname]["b"][0])
    return coin + ": " + str((kethusdask + kethusdbid) / 2)[0:6]


class Module(core.module.Module):
    @core.decorators.every(minutes=30)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.curprice))

        self.__curprice = ""
        self.__getbtc = util.format.asbool(self.parameter("getbtc", True))
        self.__geteth = util.format.asbool(self.parameter("geteth", True))
        self.__getltc = util.format.asbool(self.parameter("getltc", True))
        self.__getcur = self.parameter("getcur", "usd")
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="xdg-open https://cryptowat.ch/"
        )

    def curprice(self, widget):
        return self.__curprice

    def update(self):
        currency = self.__getcur
        btcprice, ethprice, ltcprice = "", "", ""
        if self.__getbtc:
            btcprice = getfromkrak("Btc", currency)
        if self.__geteth:
            ethprice = getfromkrak("Eth", currency)
        if self.__getltc:
            ltcprice = getfromkrak("Ltc", currency)
        self.__curprice = (
            btcprice
            + " " * (self.__getbtc * self.__geteth)
            + ethprice
            + " " * (self.__getltc * max(self.__getbtc, self.__geteth))
            + ltcprice
        )


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Displays the current song being played and allows pausing, skipping ahead, and skipping back.

Requires the following library:
    * python-dbus

Parameters:
    * spotify.format:   Format string (defaults to '{artist} - {title}')
      Available values are: {album}, {title}, {artist}, {trackNumber}
    * spotify.layout:   Comma-separated list to change order of widgets (defaults to song, previous, pause, next)
      Widget names are: spotify.song, spotify.prev, spotify.pause, spotify.next

contributed by `yvesh <https://github.com/yvesh>`_ - many thanks!

added controls by `LtPeriwinkle <https://github.com/LtPeriwinkle>`_ - many thanks!
"""

import sys
import dbus

import core.module
import core.widget
import core.input
import core.decorators
import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        self.__layout = self.parameter(
            "layout",
            util.format.aslist("spotify.song,spotify.prev,spotify.pause,spotify.next"),
        )

        self.__song = ""
        self.__pause = ""
        self.__format = self.parameter("format", "{artist} - {title}")

        self.__cmd = "dbus-send --session --type=method_call --dest=org.mpris.MediaPlayer2.spotify \
                /org/mpris/MediaPlayer2 org.mpris.MediaPlayer2.Player."

    def hidden(self):
        return self.string_song == ""

    def __get_song(self):
        bus = dbus.SessionBus()
        spotify = bus.get_object(
            "org.mpris.MediaPlayer2.spotify", "/org/mpris/MediaPlayer2"
        )
        spotify_iface = dbus.Interface(spotify, "org.freedesktop.DBus.Properties")
        props = spotify_iface.Get("org.mpris.MediaPlayer2.Player", "Metadata")
        playback_status = str(
            spotify_iface.Get("org.mpris.MediaPlayer2.Player", "PlaybackStatus")
        )
        if playback_status == "Playing":
            self.__pause = "\u258D\u258D"
        else:
            self.__pause = "\u25B6"
            self.__song = self.__format.format(
            album=str(props.get("xesam:album")),
            title=str(props.get("xesam:title")),
                artist=",".join(props.get("xesam:artist")),
                trackNumber=str(props.get("xesam:trackNumber")),
            )

    def update(self):
        try:
            self.clear_widgets()
            self.__get_song()
            
            widget_map = {}
            for widget_name in self.__layout:
                widget = self.add_widget(name=widget_name)
                if widget_name == "spotify.prev":
                    widget_map[widget] = {
                        "button": core.input.LEFT_MOUSE,
                        "cmd": self.__cmd + "Previous",
                    }
                    widget.full_text("\u258F\u25C0")
                elif widget_name == "spotify.pause":
                    widget_map[widget] = {
                        "button": core.input.LEFT_MOUSE,
                        "cmd": self.__cmd + "PlayPause",
                    }
                    widget.full_text(self.__pause)
                elif widget_name == "spotify.next":
                    widget_map[widget] = {
                        "button": core.input.LEFT_MOUSE,
                        "cmd": self.__cmd + "Next",
                    }
                    widget.full_text("\u25B6\u2595")
                elif widget_name == "spotify.song":
                    widget.full_text(self.__song)
                else:
                    raise KeyError(
                        "The spotify module does not have a {widget_name!r} widget".format(
                            widget_name=widget_name
                        )
                    )
            for widget, callback_options in widget_map.items():
                core.input.register(widget, **callback_options)

        except Exception:
            self.__song = ""

    @property
    def string_song(self):
        if sys.version_info.major < 3:
            return unicode(self.__song)
        return str(self.__song)



# pylint: disable=C0111,R0903

"""Displays focused i3 window title.

Requirements:
    * i3ipc

Parameters:
    * title.max : Maximum character length for title before truncating. Defaults to 64.
    * title.placeholder : Placeholder text to be placed if title was truncated. Defaults to '...'.
    * title.scroll : Boolean flag for scrolling title. Defaults to False


contributed by `UltimatePancake <https://github.com/UltimatePancake>`_ - many thanks!
"""

import threading

try:
    import i3ipc
except ImportError:
    pass

no_title = "n/a"

import core.module
import core.decorators

import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, [])

        # parsing of parameters
        self.__scroll = util.format.asbool(self.parameter("scroll", False))
        self.__max = int(self.parameter("max", 64))
        self.__placeholder = self.parameter("placeholder", "...")
        self.__title = ""

        # set output of the module
        self.add_widget(
            full_text=self.__scrolling_focused_title
            if self.__scroll
            else self.__focused_title
        )

        # create a connection with i3ipc
        self.__i3 = i3ipc.Connection()
        # event is called both on focus change and title change
        self.__i3.on("window", lambda __p_i3, __p_e: self.__pollTitle())
        # begin listening for events
        threading.Thread(target=self.__i3.main).start()

        # initialize the first title
        self.__pollTitle()

    def __focused_title(self, widget):
        return self.__title

    @core.decorators.scrollable
    def __scrolling_focused_title(self, widget):
        return self.__full_title

    def __pollTitle(self):
        """Updating current title."""
        try:
            self.__full_title = self.__i3.get_tree().find_focused().name
        except:
            self.__full_title = no_title
        if self.__full_title is None:
            self.__full_title = no_title

        if not self.__scroll:
            # cut the text if it is too long
            if len(self.__full_title) > self.__max:
                self.__title = self.__full_title[
                    0 : self.__max - len(self.__placeholder)
                ]
                self.__title = "{}{}".format(self.__title, self.__placeholder)
            else:
                self.__title = self.__full_title


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



"""Pop-up menus."""

import logging

import tkinter as tk

import functools


class menu(object):
    """Draws a hierarchical popup menu

    :param parent: If given, this menu is a leave of the "parent" menu
    :param leave: If set to True, close this menu when mouse leaves the area (defaults to True)
    """

    def __init__(self, parent=None, leave=True):
        self.running = True
        self.parent = None
        if not parent:
            self._root = tk.Tk()
            self._root.withdraw()
            self._menu = tk.Menu(self._root, tearoff=0)
            self._menu.bind("<FocusOut>", self.__on_focus_out)

            self.add_menuitem("close", self.__on_focus_out)
            self.add_separator()
        else:
            self._root = parent.root()
            self._root.withdraw()
            self._menu = tk.Menu(self._root, tearoff=0)
            self._menu.bind("<FocusOut>", self.__on_focus_out)
            self.parent = parent
        if leave:
            self._menu.bind("<Leave>", self.__on_focus_out)

        self._menu.bind("<ButtonRelease-1>", self.release)

    """Returns the root node of this menu

    :return: root node
    """

    def root(self):
        return self._root

    """Returns the menu

    :return: menu
    """

    def menu(self):
        return self._menu

    def __on_focus_out(self, event=None):
        self._root.destroy()

    def __on_click(self, callback):
        self._root.destroy()
        callback()

    def release(self, event=None):
        self.running=False
        if self.parent:
            self.parent.release(event)

    """Adds a cascading submenu to the current menu

    :param menuitem: label to display for the submenu
    :param submenu: submenu to show
    """

    def add_cascade(self, menuitem, submenu):
        self._menu.add_cascade(label=menuitem, menu=submenu.menu())

    """Adds an item to the current menu

    :param menuitem: label to display for the entry
    :param callback: method to invoke on click
    """

    def add_menuitem(self, menuitem, callback):
        self._menu.add_command(
            label=menuitem, command=functools.partial(self.__on_click, callback)
        )

    """Adds a separator to the menu in the current location"""

    def add_separator(self):
        self._menu.add_separator()

    """Shows this menu

    :param event: i3wm event that triggered the menu (dict that contains "x" and "y" fields)
    :param offset_x: x-axis offset from mouse position for the menu (defaults to 0)
    :param offset_y: y-axis offset from mouse position for the menu (defaults to 0)
    """

    def show(self, event, offset_x=0, offset_y=0):
        try:
            self._menu.tk_popup(event["x"] + offset_x, event["y"] + offset_y)
        finally:
            self._menu.grab_release()

        while self.running == True:
            try:
                self._root.update_idletasks()
                self._root.update()
            except:
                self.running = False
        try:
            self._root.destroy()
        except:
            pass


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



import os
import shlex
import subprocess
import logging


def execute(
    cmd,
    wait=True,
    ignore_errors=False,
    include_stderr=False,
    env=None,
    return_exitcode=False,
    shell=False,
):
    """Executes a commandline utility and returns its output

    :param cmd: the command (as string) to execute, returns the program's output
    :param wait: set to True to wait for command completion, False to return immediately, defaults to True
    :param ignore_errors: set to True to return a string when an exception is thrown, otherwise might throw, defaults to False
    :param include_stderr: set to True to include stderr output in the return value, defaults to False
    :param env: provide a dict here to specify a custom execution environment, defaults to None
    :param return_exitcode: set to True to return a pair, where the first member is the exit code and the message the second, defaults to False
    :param shell: set to True to run command in a separate shell, defaults to False

    :raises RuntimeError: the command either didn't exist or didn't exit cleanly, and ignore_errors was set to False

    :return: output of cmd, or stderr, if ignore_errors is True and the command failed; or a tuple of exitcode and the previous, if return_exitcode is set to True
    :rtype: string or tuple (if return_exitcode is set to True)
    """
    args = cmd if shell else shlex.split(cmd)
    logging.debug(cmd)

    if not env:
        env = os.environ.copy()

    myenv = env.copy()

    myenv["LC_ALL"] = "C"
    if "WAYLAND_SOCKET" in myenv:
        del myenv["WAYLAND_SOCKET"]

    try:
        proc = subprocess.Popen(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT if include_stderr else subprocess.PIPE,
            env=myenv,
            shell=shell,
        )
    except FileNotFoundError as e:
        raise RuntimeError("{} not found".format(cmd))

    if wait:
        out, _ = proc.communicate()
        if proc.returncode != 0:
            err = "{} exited with code {}".format(cmd, proc.returncode)
            logging.warning(err)
            if ignore_errors:
                return (proc.returncode, err) if return_exitcode else err
            raise RuntimeError(err)
        res = out.decode("utf-8")
        logging.debug(res)
        return (proc.returncode, res) if return_exitcode else res
    return (0, "") if return_exitcode else ""


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903

"""Displays available RAM, total amount of RAM and percentage available.

Parameters:
    * memory.warning : Warning threshold in % of memory used (defaults to 80%)
    * memory.critical: Critical threshold in % of memory used (defaults to 90%)
    * memory.format: Format string (defaults to '{used}/{total} ({percent:05.02f}%)')
    * memory.usedonly: Only show the amount of RAM in use (defaults to False). Same as memory.format='{used}'
"""

import re

import core.module
import core.widget
import core.input

import util.format


class Module(core.module.Module):
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(self.memory_usage))
        core.input.register(
            self, button=core.input.LEFT_MOUSE, cmd="gnome-system-monitor"
        )

    @property
    def _format(self):
        if util.format.asbool(self.parameter("usedonly", False)):
            return "{used}"
        else:
            return self.parameter("format", "{used}/{total} ({percent:05.02f}%)")

    def memory_usage(self, widget):
        return self._format.format(**self._mem)

    def update(self):
        data = {}
        with open("/proc/meminfo", "r") as f:
            for line in f:
                tmp = re.split(r"[:\s]+", line)
                value = int(tmp[1])
                if tmp[2] == "kB":
                    value = value * 1024
                if tmp[2] == "mB":
                    value = value * 1024 * 1024
                if tmp[2] == "gB":
                    value = value * 1024 * 1024 * 1024
                data[tmp[0]] = value
        if "MemAvailable" in data:
            used = data["MemTotal"] - data["MemAvailable"]
        else:
            used = (
                data["MemTotal"]
                - data["MemFree"]
                - data["Buffers"]
                - data["Cached"]
                - data["Slab"]
            )
        self._mem = {
            "total": util.format.byte(data["MemTotal"]),
            "available": util.format.byte(data["MemAvailable"]),
            "free": util.format.byte(data["MemFree"]),
            "used": util.format.byte(used),
            "percent": float(used) / float(data["MemTotal"]) * 100.0,
        }

    def state(self, widget):
        if self._mem["percent"] > float(self.parameter("critical", 90)):
            return "critical"
        if self._mem["percent"] > float(self.parameter("warning", 80)):
            return "warning"
        return None


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4



# pylint: disable=C0111,R0903,W0212

"""Enable/disable automatic screen locking.

Requires the following executables:
    * xdg-screensaver
    * xdotool
    * xprop (as dependency for xdotool)
    * notify-send

contributed by `TheEdgeOfRage <https://github.com/TheEdgeOfRage>`_ - many thanks!
"""

import logging
import os
import shutil
import psutil

import core.module
import core.widget
import core.input
import core.decorators

import util.cli


class Module(core.module.Module):
    @core.decorators.every(minutes=10)
    def __init__(self, config, theme):
        super().__init__(config, theme, core.widget.Widget(""))

        self.__active = False
        self.__xid = None

        core.input.register(self, button=core.input.LEFT_MOUSE, cmd=self.__toggle)

    def __check_requirements(self):
        requirements = ["xdotool", "xprop", "xdg-screensaver"]
        missing = []
        for tool in requirements:
            if not shutil.which(tool):
                missing.append(tool)
        return missing

    def __get_i3bar_xid(self):
        xid = (
            util.cli.execute("xdotool search --class 'i3bar'")
            .partition("\n")[0]
            .strip()
        )
        if xid.isdigit():
            return xid
        logging.warning("Module caffeine: xdotool couldn't get X window ID of 'i3bar'.")
        return None

    def __notify(self):
        if not shutil.which("notify-send"):
            return

        if self.__active:
            util.cli.execute("notify-send 'Consuming caffeine'")
        else:
            util.cli.execute("notify-send 'Out of coffee'")

    def _suspend_screensaver(self):
        self.__xid = self.__get_i3bar_xid()
        if self.__xid is None:
            return False

        pid = os.fork()
        if pid == 0:
            os.setsid()
            util.cli.execute("xdg-screensaver suspend {}".format(self.__xid))
            os._exit(0)
        else:
            os.waitpid(pid, 0)
        return True

    def __resume_screensaver(self):
        success = True
        xprop_path = shutil.which("xprop")
        pids = [
            p.pid
            for p in psutil.process_iter()
            if p.cmdline() == [xprop_path, "-id", str(self.__xid), "-spy"]
        ]
        for pid in pids:
            try:
                os.kill(pid, 9)
            except OSError:
                success = False
        return success

    def state(self, _):
        if self.__active:
            return "activated"
        return "deactivated"

    def __toggle(self, _):
        missing = self.__check_requirements()
        if missing:
            logging.warning("Could not run caffeine - missing %s!", ", ".join(missing))
            return

        self.__active = not self.__active
        if self.__active:
            success = self._suspend_screensaver()
        else:
            success = self.__resume_screensaver()

        if success:
            self.__notify()
        else:
            self.__active = not self.__active


# vim: tabstop=8 expandtab shiftwidth=4 softtabstop=4
